#include "../include/standard.h" 
#include "../include/typedefs.h"
#include "../include/definitions.h"
#include "../include/utillib.h"
#include "../include/cmalloc.h"

#define DEBUG
#undef DEBUG


/*========================================================================*/
/*========================================================================*/
void Bend::setup(const Pot_data_entry *database, const Maps *maps, 
                                                 const Pos *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/

  int ires, iresbond, ibend_global;
  int i, iii, ibend, jbend, n; 
  int bend_check;
  int atoms[4];

  double fk_max;
  double *x, *y, *z;
  double dx21, dy21, dz21, r21, rr21;
  double dx23, dy23, dz23, r23, rr23;
  double dx31, dy31, dz31, r31, rr31, dr31;
  double cost, sint, theta, dtheta, vbend;

  extern System *global_system;
  System *system = global_system;

  const Pot_data_entry *database_entry, *exact_match_entry, *wild_match_entry;
  char   err_msg[LINE_NCHARMAX];

  printf("----------------------------------------------\n");
  printf("Setting up bends\n...\n");  

/*-------------------------------------------------------------------------*/
/* I) Setup topology                                                       */
  
  nbend = maps->nbend;
  atom1 = new int[nbend]-1;
  atom2 = new int[nbend]-1;
  atom3 = new int[nbend]-1;
  fk_bend = new double[nbend]-1;
  theta0  = new double[nbend]-1;
  fk_bond = new double[nbend]-1;
  r0      = new double[nbend]-1;

  for (ires=1;ires<=pos->nres;ires++) {
      for (ibend=1;ibend<=pos->residue[ires].nbend;ibend++) {
   	  ibend_global = pos->residue[ires].bend[ibend];
          if (ibend_global>nbend) {
             sprintf(err_msg,"Allocation problem in bend setup topology!");
             sprintf(err_msg,"ibend_global:%d>nbend %d",ibend_global,nbend);
             errorMessage(err_msg);
          } 
   	  atom1[ibend_global] = pos->residue[ires].bend_atom[ibend][1];
   	  atom2[ibend_global] = pos->residue[ires].bend_atom[ibend][2];
   	  atom3[ibend_global] = pos->residue[ires].bend_atom[ibend][3];
      }/*end for*/       
  }/*end for*/
  for (iresbond=1;iresbond<=maps->nresbond;iresbond++) {
      for (ibend=1;ibend<=pos->residue_bond[iresbond].nbend;ibend++) {
   	  ibend_global = pos->residue_bond[iresbond].bend[ibend];
          if (ibend_global>nbend) {
             sprintf(err_msg,"Allocation problem in bend setup topology!");
             sprintf(err_msg,"ibend_global:%d>nbend %d",ibend_global,nbend);
             errorMessage(err_msg);
          } 
   	  atom1[ibend_global] = pos->residue_bond[iresbond].bend_atom[ibend][1];
   	  atom2[ibend_global] = pos->residue_bond[iresbond].bend_atom[ibend][2];
   	  atom3[ibend_global] = pos->residue_bond[iresbond].bend_atom[ibend][3];
      }/*end for*/
  }/*end for*/

  for (ibend=1; ibend<=nbend; ibend++) {
    if (atom1[ibend]==atom2[ibend] ||
	atom1[ibend]==atom3[ibend] ||
	atom2[ibend]==atom3[ibend] ) {
      sprintf(err_msg, "Ilegal bend #%d: same atom index.", ibend);
      errorMessage(err_msg);
    }
  }

  for (ibend=1; ibend<=nbend; ibend++) {
    for (jbend=ibend+1; jbend<=nbend; jbend++) {
      if (atom1[ibend]==atom1[jbend] &&
	  atom2[ibend]==atom2[jbend] &&
	  atom3[ibend]==atom3[jbend]   ){
	sprintf(err_msg, "Bends %d and %d repeat.", ibend, jbend);
	errorMessage(err_msg);
      }
      if (atom1[ibend]==atom3[jbend] &&
	  atom2[ibend]==atom2[jbend] &&
	  atom3[ibend]==atom1[jbend]   ){
	sprintf(err_msg, "Bends %d and %d repeat.", ibend, jbend);
	errorMessage(err_msg);
      }
    }
  }

/*-------------------------------------------------------------------------*/
/* II) Setup potential parameters                                          */

  for (ibend=1; ibend<=nbend; ibend++) {
    atoms[1] = atom1[ibend];
    atoms[2] = atom2[ibend];
    atoms[3] = atom3[ibend];
    exact_match_entry = NULL;
    wild_match_entry  = NULL;
    database_entry    = database;
    bend_check = 0;
    while (database_entry!=NULL && bend_check!=1) {
      bend_check = database_entry->match(maps, atoms, 2);
      if (bend_check==1) { exact_match_entry = database_entry; }
      if (bend_check==2) { wild_match_entry  = database_entry; }
      database_entry = database_entry->next;
    }/*end while */
    if (exact_match_entry==NULL && wild_match_entry==NULL) {
      sprintf(err_msg, "Bend is not found in database.\nindices %d %d %d \nlabels %s %s %s",
	      atoms[1], atoms[2], atoms[3],
              maps->atm_label[atoms[1]],
              maps->atm_label[atoms[2]],
              maps->atm_label[atoms[3]]);
      errorMessage(err_msg);
    }
    if (exact_match_entry!=NULL) { database_entry = exact_match_entry;}
    else                         { database_entry = wild_match_entry; }
    fk_bend[ibend] = database_entry->fk_bend;
    theta0[ibend]  = database_entry->theta0;
    fk_bond[ibend] = database_entry->fk_bond;
    r0[ibend]      = database_entry->r0;
  }/*end for: ibend */

/*-------------------------------------------------------------------------*/
/* III) In case of KB model: initialize theta0 from coordinates            */

  if (pos->cgres_model_KB) {
     x  = pos->x;
     y  = pos->y;
     z  = pos->z;
     for (ibend=1; ibend<=nbend; ibend++) {
       dx21 = x[atom1[ibend]] - x[atom2[ibend]];
       dy21 = y[atom1[ibend]] - y[atom2[ibend]];
       dz21 = z[atom1[ibend]] - z[atom2[ibend]];
       rr21 = dx21*dx21+dy21*dy21+dz21*dz21;
       r21  = sqrt(rr21);
       dx23 = x[atom3[ibend]] - x[atom2[ibend]];
       dy23 = y[atom3[ibend]] - y[atom2[ibend]];
       dz23 = z[atom3[ibend]] - z[atom2[ibend]];
       rr23 = dx23*dx23+dy23*dy23+dz23*dz23;
       r23  = sqrt(rr23);
       cost = (dx21*dx23 + dy21*dy23 + dz21*dz23)/(r21*r23);
       if (cost> 1) cost = 1;
       if (cost<-1) cost =-1;
       theta  = acos(cost);
       theta0[ibend] = theta;
     }/*end for*/
  }/*end if*/


/*-------------------------------------------------------------------------*/
/* IV) MAX force constant cutoff                                           */

  if ( (global_system->general.force_const_cutoff) > 1 ) {
        fk_max = global_system->general.force_const_cutoff_bend;
        for (ibend=1; ibend<=nbend; ibend++) {
             fk_bend[ibend] = MIN(fk_bend[ibend],fk_max);
             fk_bond[ibend] = MIN(fk_bond[ibend],fk_max);
        }/*endfor*/
  }/*endif*/

/*-------------------------------------------------------------------------*/
/* V) DEBUG                                                                */

#ifdef DEBUG
  for (ibend=1; ibend<=nbend; ibend++) {
    printf("Bend %2d:  %d(%s)--%d(%s)--%d(%s)\n", ibend,
   	   atom1[ibend], maps->atm_label[atom1[ibend]],
   	   atom2[ibend], maps->atm_label[atom2[ibend]],
   	   atom3[ibend], maps->atm_label[atom3[ibend]]);
    printf("\t\t{fk_bend=%.8g, theta0=%.8g, fk_bond=%.8g, r0=%.8g}\n",
   	   fk_bend[ibend], theta0[ibend], fk_bond[ibend], r0[ibend]);
  }
#endif  

  printf("Bends setting up done.\n");
  printf("----------------------------------------------\n");

/*-------------------------------------------------------------------------*/
} /*end routine */
/*========================================================================*/


/*========================================================================*/
/*========================================================================*/
void Bend::setup(const Pot_data_entry *database, const Maps *maps, 
		 const char *parm_filename)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/
  int i, iii, ibend, jbend, n; 
  int natm_tot, nkey, key_defined, line_num;
  int nbend_now, nbend_mall, atom1_ind, atom2_ind, atom3_ind;
  int bend_check, key_found[4];
  int atoms[4];
  const Pot_data_entry *database_entry, *exact_match_entry, *wild_match_entry;
  
  char   err_msg[LINE_NCHARMAX];
  char entry_line[LINE_NCHARMAX];
  char entry_name[NCHARMAX], key[NCHARMAX], arg[NCHARMAX];
  char *def;
  FILE *parm_file;
  
  char *bend_def_keys[] = {"", "atom1", "atom2", "atom3"};

  printf("----------------------------------------------\n");
  printf("Setting up bends\n...\n");

  if ((parm_file=fopen(parm_filename, "r")) == NULL) {
    sprintf(err_msg, "Cannot open parm file %s.", parm_filename);
    errorMessage(err_msg);
  }
  
  nbend_mall= NMEM_MIN;
  nbend_now = 0;
  natm_tot  = maps->natm_tot;
  line_num  = 0;
  
  atom1 = new int[nbend_mall]-1;
  atom2 = new int[nbend_mall]-1;
  atom3 = new int[nbend_mall]-1;
  
  while((n=readOneEntry(parm_file,'~','[',']',entry_line, line_num))!=0) {
    if (n!=1) {
      sprintf(err_msg, "%s at line %d while reading file %s.",
	      (n==-1) ? "Syntax error" : "Internal error", 
	      line_num, parm_filename);
      errorMessage(err_msg);
    }
    n=getEntryName(entry_line, '~', '[', entry_name, def);
    if (n!=1) {
      sprintf(err_msg, "Syntax error at parm file line:\n%s", entry_line);
      errorMessage(err_msg);
    }
    if (strcasecmp(entry_name, "bend_def")==0) {
      nkey = 3;
      for ( i=1; i<=nkey; i++) { key_found[i] = 0; }
      while ((n=getKeyArgPair(def, key, arg, '\\', '{', '}', def))==1) {
	key_defined = 0;
	for ( i=1; i<=nkey; i++) {                  /*loop for key words.*/
	  if (strcasecmp(key, bend_def_keys[i])==0) {/*it's a keyword.  */
	    key_defined  = 1;
	    key_found[i] = 1;
	    switch (i) {
	    case 1:
	      sscanf(arg, "%d", &atom1_ind);
	      break;
	    case 2:
	      sscanf(arg, "%d", &atom2_ind);
	      break;
	    case 3:
	      sscanf(arg, "%d", &atom3_ind);
	      break;
	    }/*switch */
	  }
	}/*end for*/
	if (key_defined==0) {
	  sprintf(err_msg, "Key word %s unknown in parmfile line:\n%s",
		  key, entry_line);
	  errorMessage(err_msg);
	}
      }/*end while: getKeyArgPair*/
      if (n!=0) {
	sprintf(err_msg, "Syntax error in parm file line:\n%s",entry_line);
	errorMessage(err_msg);
      }
      for ( i=1; i<=nkey; i++) { 
	if (key_found[i]==0) {
	  sprintf(err_msg, "Key word %s missed in parm file line:\n%s",
		  bend_def_keys[i], entry_line);
	  errorMessage(err_msg);
	}
      }
      if (atom1_ind<1 || atom1_ind>natm_tot || 
	  atom2_ind<1 || atom2_ind>natm_tot || 
	  atom3_ind<1 || atom3_ind>natm_tot    ) {
	sprintf(err_msg, "BEND::setup(): Illegal atom index in parm file %s.",
		parm_filename);
	errorMessage(err_msg);
      }
      nbend_now++;
      if (nbend_now>nbend_mall) {
	nbend_mall += NMEM_MIN;
	atom1       = (int *) crealloc(atom1+1, nbend_mall*sizeof(int))-1;
	atom2       = (int *) crealloc(atom2+1, nbend_mall*sizeof(int))-1;
	atom3       = (int *) crealloc(atom3+1, nbend_mall*sizeof(int))-1;
      }
      atom1[nbend_now] = atom1_ind;
      atom2[nbend_now] = atom2_ind;
      atom3[nbend_now] = atom3_ind;
    }/*end if: bend_def */
  }/*end while: readOneEntry*/
  fclose(parm_file);
  
  nbend   = nbend_now;
  fk_bend = new double[nbend]-1;
  theta0  = new double[nbend]-1;
  fk_bond = new double[nbend]-1;
  r0      = new double[nbend]-1;

  for (ibend=1; ibend<=nbend; ibend++) {
    atoms[1] = atom1[ibend];
    atoms[2] = atom2[ibend];
    atoms[3] = atom3[ibend];
    exact_match_entry = NULL;
    wild_match_entry  = NULL;
    database_entry    = database;
    bend_check = 0;
    while (database_entry!=NULL && bend_check!=1) {
      bend_check = database_entry->match(maps, atoms, 2);
      if (bend_check==1) { exact_match_entry = database_entry; }
      if (bend_check==2) { wild_match_entry  = database_entry; }
      database_entry = database_entry->next;
    }/*end while */
    if (exact_match_entry==NULL && wild_match_entry==NULL) {
      sprintf(err_msg, "Bend is not found in database.\nindices %d %d %d \nlabels %s %s %s",
	      atoms[1], atoms[2], atoms[3],
              maps->atm_label[atoms[1]],
              maps->atm_label[atoms[2]],
              maps->atm_label[atoms[3]]);
      errorMessage(err_msg);
    }
    if (exact_match_entry!=NULL) { database_entry = exact_match_entry;}
    else                         { database_entry = wild_match_entry; }
    fk_bend[ibend] = database_entry->fk_bend;
    theta0[ibend]  = database_entry->theta0;
    fk_bond[ibend] = database_entry->fk_bond;
    r0[ibend]      = database_entry->r0;
  }/*end for: ibend */

  for (ibend=1; ibend<=nbend; ibend++) {
    if (atom1[ibend]==atom2[ibend] ||
	atom1[ibend]==atom3[ibend] ||
	atom2[ibend]==atom3[ibend] ) {
      sprintf(err_msg, "Ilegal bend #%d: same atom index.", ibend);
      errorMessage(err_msg);
    }
  }

  for (ibend=1; ibend<=nbend; ibend++) {
    for (jbend=ibend+1; jbend<=nbend; jbend++) {
      if (atom1[ibend]==atom1[jbend] &&
	  atom2[ibend]==atom2[jbend] &&
	  atom3[ibend]==atom3[jbend]   ){
	sprintf(err_msg, "Bends %d and %d repeat.", ibend, jbend);
	errorMessage(err_msg);
      }
      if (atom1[ibend]==atom3[jbend] &&
	  atom2[ibend]==atom2[jbend] &&
	  atom3[ibend]==atom1[jbend]   ){
	sprintf(err_msg, "Bends %d and %d repeat.", ibend, jbend);
	errorMessage(err_msg);
      }
    }
  }

#ifdef DEBUG
  for (ibend=1; ibend<=nbend; ibend++) {
    printf("Bend %2d:  %d(%s)--%d(%s)--%d(%s)\n", ibend,
   	   atom1[ibend], maps->atm_label[atom1[ibend]],
   	   atom2[ibend], maps->atm_label[atom2[ibend]],
   	   atom3[ibend], maps->atm_label[atom3[ibend]]);
    printf("\t\t{fk_bend=%.8g, theta0=%.8g, fk_bond=%.8g, r0=%.8g}\n",
   	   fk_bend[ibend], theta0[ibend], fk_bond[ibend], r0[ibend]);
  }
#endif  

  printf("Bends setting up done.\n");
  printf("----------------------------------------------\n");

/*-------------------------------------------------------------------------*/
} /*end routine */
/*========================================================================*/



/*========================================================================*/
/*========================================================================*/
double Bend::energy(const Pos *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/
  int i, ibend;
  int atom1_ind, atom2_ind, atom3_ind;

  double *x, *y, *z, *fx, *fy, *fz;
  double dx21, dy21, dz21, r21, rr21;
  double dx23, dy23, dz23, r23, rr23;
  double dx31, dy31, dz31, r31, rr31, dr31;
  double rq1x, rq1y, rq1z, rq2x, rq2y, rq2z;
  double cost, sint, theta, dtheta, pre, vbend_vbond;
  
  x  = pos->x;
  y  = pos->y;
  z  = pos->z;
  fx = pos->fx;
  fy = pos->fy;
  fz = pos->fz;

/*---------------------------------------------------------------------------*/
/* 0) Return if no bend term                                                 */

  if (energy_term != 1) { 
      vbend = 0.0;
      vbond = 0.0;
      return (vbend+vbond);
  }/*endif*/

/*--------------------------------------------------------------------------*/
/* I) Bend part                                                             */
  vbend = 0.0;
  for (ibend=1; ibend<=nbend; ibend++) {
    atom1_ind = atom1[ibend];
    atom2_ind = atom2[ibend];
    atom3_ind = atom3[ibend];
    dx21 = x[atom1_ind] - x[atom2_ind];
    dy21 = y[atom1_ind] - y[atom2_ind];
    dz21 = z[atom1_ind] - z[atom2_ind];
    rr21 = dx21*dx21+dy21*dy21+dz21*dz21;
    r21  = sqrt(rr21);
    dx23 = x[atom3_ind] - x[atom2_ind];
    dy23 = y[atom3_ind] - y[atom2_ind];
    dz23 = z[atom3_ind] - z[atom2_ind];
    rr23 = dx23*dx23+dy23*dy23+dz23*dz23;
    r23  = sqrt(rr23);
    cost = (dx21*dx23 + dy21*dy23 + dz21*dz23)/(r21*r23);
    if (cost> 1) cost = 1;
    if (cost<-1) cost =-1;
    sint   = sqrt(1-cost*cost);
    theta  = acos(cost);
    dtheta = theta-theta0[ibend];
    vbend += 0.5*fk_bend[ibend]*dtheta*dtheta;
    
    rq1x= dx23 - cost*dx21*r23/r21;
    rq1y= dy23 - cost*dy21*r23/r21;
    rq1z= dz23 - cost*dz21*r23/r21;
    rq2x= dx21 - cost*dx23*r21/r23;
    rq2y= dy21 - cost*dy23*r21/r23;
    rq2z= dz21 - cost*dz23*r21/r23;
    pre = fk_bend[ibend]*dtheta/(sint*r21*r23);
    
    fx[atom1_ind] += pre*rq1x;
    fy[atom1_ind] += pre*rq1y;
    fz[atom1_ind] += pre*rq1z;
    fx[atom2_ind] -= (pre*rq1x+pre*rq2x);
    fy[atom2_ind] -= (pre*rq1y+pre*rq2y);
    fz[atom2_ind] -= (pre*rq1z+pre*rq2z);
    fx[atom3_ind] += pre*rq2x;
    fy[atom3_ind] += pre*rq2y;
    fz[atom3_ind] += pre*rq2z;
  }/*end for*/
  
/*--------------------------------------------------------------------------*/
/* II) Bond part                                                            */
  vbond = 0.0;
  for (ibend=1; ibend<=nbend; ibend++) {
    atom1_ind = atom1[ibend];
    atom2_ind = atom2[ibend];
    atom3_ind = atom3[ibend];
    dx31 = x[atom1_ind] - x[atom3_ind];
    dy31 = y[atom1_ind] - y[atom3_ind];
    dz31 = z[atom1_ind] - z[atom3_ind];
    rr31 = dx31*dx31+dy31*dy31+dz31*dz31;
    r31  = sqrt(rr31);
    dr31 = r31 - r0[ibend];
    vbond += 0.5*fk_bond[ibend]*dr31*dr31;
    
    pre = -fk_bond[ibend]*dr31/r31;
    fx[atom1_ind] += pre*dx31;
    fy[atom1_ind] += pre*dy31;
    fz[atom1_ind] += pre*dz31;
    fx[atom3_ind] -= pre*dx31;
    fy[atom3_ind] -= pre*dy31;
    fz[atom3_ind] -= pre*dz31;
  }/*end for */
  vbend_vbond = vbend + vbond;
  return vbend_vbond;
  
/*-------------------------------------------------------------------------*/
} /*end routine */
/*========================================================================*/



/*========================================================================*/
/*========================================================================*/
double Bend::getPot(const Pos *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/
  int i, ibend;
  int atom1_ind, atom2_ind, atom3_ind;

  double *x, *y, *z;
  double dx21, dy21, dz21, r21, rr21;
  double dx23, dy23, dz23, r23, rr23;
  double dx31, dy31, dz31, r31, rr31, dr31;
  double cost, sint, theta, dtheta, vbend_vbond;
  
  x  = pos->x;
  y  = pos->y;
  z  = pos->z;


/*---------------------------------------------------------------------------*/
/* 0) Return if no bend term                                                 */

  if (energy_term != 1) { 
      vbend = 0.0;
      vbond = 0.0;
      return (vbend+vbond);
  }/*endif*/

/*--------------------------------------------------------------------------*/
/* I) Bend part       */
  vbend = 0.0;
  for (ibend=1; ibend<=nbend; ibend++) {
    atom1_ind = atom1[ibend];
    atom2_ind = atom2[ibend];
    atom3_ind = atom3[ibend];
    dx21 = x[atom1_ind] - x[atom2_ind];
    dy21 = y[atom1_ind] - y[atom2_ind];
    dz21 = z[atom1_ind] - z[atom2_ind];
    rr21 = dx21*dx21+dy21*dy21+dz21*dz21;
    r21  = sqrt(rr21);
    dx23 = x[atom3_ind] - x[atom2_ind];
    dy23 = y[atom3_ind] - y[atom2_ind];
    dz23 = z[atom3_ind] - z[atom2_ind];
    rr23 = dx23*dx23+dy23*dy23+dz23*dz23;
    r23  = sqrt(rr23);
    cost = (dx21*dx23 + dy21*dy23 + dz21*dz23)/(r21*r23);
    if (cost> 1) cost = 1;
    if (cost<-1) cost =-1;
    theta  = acos(cost);
    dtheta = theta-theta0[ibend];
    vbend += 0.5*fk_bend[ibend]*dtheta*dtheta;
  }/*end for*/
  
/*--------------------------------------------------------------------------*/
/* II) Bond part       */
  vbond = 0.0;
  for (ibend=1; ibend<=nbend; ibend++) {
    atom1_ind = atom1[ibend];
    atom2_ind = atom2[ibend];
    atom3_ind = atom3[ibend];
    dx31 = x[atom1_ind] - x[atom3_ind];
    dy31 = y[atom1_ind] - y[atom3_ind];
    dz31 = z[atom1_ind] - z[atom3_ind];
    rr31 = dx31*dx31+dy31*dy31+dz31*dz31;
    r31  = sqrt(rr31);
    dr31 = r31 - r0[ibend];
    vbond += 0.5*fk_bond[ibend]*dr31*dr31;
    
  }/*end for */

  vbend_vbond = vbond + vbend;
  return vbend_vbond;
  
/*-------------------------------------------------------------------------*/
} /*end routine */
/*========================================================================*/
