#include "../include/standard.h" 
#include "../include/typedefs.h"
#include "../include/definitions.h"
#include "../include/utillib.h"
#include "../include/cmalloc.h"

#define DEBUG
#undef DEBUG

/*========================================================================*/
/*========================================================================*/
void Bond::setup(const Pot_data_entry *database, const Maps *maps, 
                                                 const Pos  *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/

  int ires, iresbond, ibond_global;
  int i, iii, ibond, jbond, iatm, jatm, n;
  int bond_check;
  int nbond_atm_tmp;
  int  atoms[3];

  double fk_max;
  double *x, *y, *z;
  double dx, dy, dz, r, rr, dr;

  const Pot_data_entry *database_entry, *exact_match_entry, *wild_match_entry;

  extern System *global_system;
  System *system = global_system;

  char   err_msg[LINE_NCHARMAX];


  printf("----------------------------------------------\n");
  printf("Setting up bonds\n...\n");


/*-------------------------------------------------------------------------*/
/* I) Setup topology                                                       */
 
  nbond = maps->nbond;
  atom1 = new int[nbond]-1;
  atom2 = new int[nbond]-1;
  fk    = new double[nbond]-1;
  r0    = new double[nbond]-1;

  for (ires=1;ires<=pos->nres;ires++) {
      for (ibond=1;ibond<=pos->residue[ires].nbond;ibond++) {
   	  ibond_global = pos->residue[ires].bond[ibond];
          if (ibond_global>nbond) {
             sprintf(err_msg,"Allocation problem in bond setup topology!");
             sprintf(err_msg,"ibond_global:%d>nbond %d",ibond_global,nbond);
             errorMessage(err_msg);
          } 
   	  atom1[ibond_global] = pos->residue[ires].bond_atom[ibond][1];
   	  atom2[ibond_global] = pos->residue[ires].bond_atom[ibond][2];
      }/*end for*/       
  }/*end for*/
  for (iresbond=1;iresbond<=maps->nresbond;iresbond++) {
      for (ibond=1;ibond<=pos->residue_bond[iresbond].nbond;ibond++) {
   	  ibond_global = pos->residue_bond[iresbond].bond[ibond];
          if (ibond_global>nbond) {
             sprintf(err_msg,"Allocation problem in bond setup topology!");
             sprintf(err_msg,"ibond_global:%d>nbond %d",ibond_global,nbond);
             errorMessage(err_msg);
          } 
   	  atom1[ibond_global] = pos->residue_bond[iresbond].bond_atom[ibond][1];
   	  atom2[ibond_global] = pos->residue_bond[iresbond].bond_atom[ibond][2];
      }/*end for*/
  }/*end for*/
  
  for (ibond=1; ibond<=nbond; ibond++) {
    if (atom1[ibond]==atom2[ibond]) {
      sprintf(err_msg, "Ilegal bond #%d: same atom index.", ibond);
      errorMessage(err_msg);
    }
  }

/*-------------------------------------------------------------------------*/
/* II) Setup potential parameters                                          */

  for (ibond=1; ibond<=nbond; ibond++) {
    atoms[1] = atom1[ibond];
    atoms[2] = atom2[ibond];
    exact_match_entry = NULL;
    wild_match_entry  = NULL;
    database_entry    = database;
    bond_check = 0;
    while (database_entry!=NULL && bond_check!=1) {
      bond_check = database_entry->match(maps, atoms, 1);
      if (bond_check==1) { exact_match_entry = database_entry; }
      if (bond_check==2) { wild_match_entry  = database_entry; }
      database_entry = database_entry->next;
    }/*end while */
    if (exact_match_entry==NULL && wild_match_entry==NULL) {
      sprintf(err_msg, "Bond is not found in database.\nindices %d %d \nlabels %s %s ",
	      atoms[1], atoms[2],
              maps->atm_label[atoms[1]],maps->atm_label[atoms[2]]);
      errorMessage(err_msg);
    }
    if (exact_match_entry!=NULL) { database_entry = exact_match_entry;}
    else                         { database_entry = wild_match_entry; }
    fk[ibond] = database_entry->fk;
    r0[ibond] = database_entry->r0;
  }/*end for: ibond */
  
  for (ibond=1; ibond<=nbond; ibond++) {
    for (jbond=ibond+1; jbond<=nbond; jbond++) {
      if (atom1[ibond]==atom1[jbond] &&
	  atom2[ibond]==atom2[jbond] ) {
	sprintf(err_msg, "Bonds %d and %d repeat.", ibond, jbond);
	errorMessage(err_msg);
      }
      if (atom1[ibond]==atom2[jbond] &&
	  atom2[ibond]==atom1[jbond] ) {
	sprintf(err_msg, "Bonds %d and %d repeat.", ibond, jbond);
	errorMessage(err_msg);
      }
    }
  }

/*-------------------------------------------------------------------------*/
/* III) Setup the atom bond map                                            */

  natm_tot = pos->natom;
  nbond_atm = new int[natm_tot]-1;
  for (iatm=1; iatm<=natm_tot; iatm++) { nbond_atm[iatm] = 0; }
  for (ibond=1; ibond<=nbond; ibond++) {
    iatm = atom1[ibond];
    jatm = atom2[ibond];
    nbond_atm[iatm]++;
    nbond_atm[jatm]++;
  }
  nbond_atm_max = 0;
  for (iatm=1; iatm<=natm_tot; iatm++) {
    if (nbond_atm[iatm]>nbond_atm_max) { nbond_atm_max = nbond_atm[iatm];}
  }
  bond_atm_map = cmall_int_mat(1, natm_tot, 1, nbond_atm_max);
  for (iatm=1; iatm<=natm_tot; iatm++) {
    nbond_atm_tmp = 0;
    for (ibond=1; ibond<=nbond; ibond++) {
      if (iatm==atom1[ibond]) {
	bond_atm_map[iatm][++nbond_atm_tmp] = atom2[ibond];
      }
      if (iatm==atom2[ibond]) {
	bond_atm_map[iatm][++nbond_atm_tmp] = atom1[ibond];
      }
    }
    if (nbond_atm_tmp!=nbond_atm[iatm]) {
      sprintf(err_msg, "Internal error of setting up atom bond map.");
      errorMessage(err_msg);
    }
  }

/*-------------------------------------------------------------------------*/
/* IV) In case of KB model: initialize r0 from coordinates                 */

  if (pos->cgres_model_KB) {
     x  = pos->x;
     y  = pos->y;
     z  = pos->z;
     for (ibond=1; ibond<=nbond; ibond++) {
         dx = x[atom1[ibond]] - x[atom2[ibond]];
         dy = y[atom1[ibond]] - y[atom2[ibond]];
         dz = z[atom1[ibond]] - z[atom2[ibond]];
         rr = dx*dx+dy*dy+dz*dz;
         r  = sqrt(rr);
         r0[ibond] = r;
     }/*end for*/
  }/*end if*/  

/*-------------------------------------------------------------------------*/
/* V) MAX force constant cutoff                                            */

  if ( (global_system->general.force_const_cutoff) > 0 ) {
        fk_max = global_system->general.force_const_cutoff_bond;
        for (ibond=1; ibond<=nbond; ibond++) 
             {fk[ibond] = MIN(fk[ibond],fk_max);}
  }/*endif*/

/*-------------------------------------------------------------------------*/
/* VI) DEBUG                                                               */

#ifdef DEBUG      
  for (ibond=1; ibond<=nbond; ibond++) {
    printf("Bond %2d:  %d(%s)--%d(%s), {fk=%.8g, r0=%.8g}\n",ibond, 
   	   atom1[ibond], maps->atm_label[atom1[ibond]],
   	   atom2[ibond], maps->atm_label[atom2[ibond]],
  	   fk[ibond], r0[ibond]);
  }
#endif
  
  printf("Bonds setting up done.\n");
  printf("----------------------------------------------\n");

/*-------------------------------------------------------------------------*/
} /*end routine */
/*=========================================================================*/


/*========================================================================*/
/*========================================================================*/
void Bond::setup(const Pot_data_entry *database, const Maps *maps, 
		 const char *parm_filename)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/
  int i, iii, ibond, jbond, iatm, jatm, n;
  int natm_tot, nkey, key_defined, line_num;
  int nbond_now, nbond_mall, atom1_ind, atom2_ind;
  int bond_check, key_found[4];
  int nbond_atm_max, nbond_atm_tmp;
  int  atoms[3];
  const Pot_data_entry *database_entry, *exact_match_entry, *wild_match_entry;
  
  char   err_msg[LINE_NCHARMAX];
  char entry_line[LINE_NCHARMAX];
  char entry_name[NCHARMAX], key[NCHARMAX], arg[NCHARMAX];
  char *def;
  FILE *parm_file;
  
  char *bond_def_keys[] = {"", "atom1", "atom2"};

  printf("----------------------------------------------\n");
  printf("Setting up bonds\n...\n");
  
  if ((parm_file=fopen(parm_filename, "r")) == NULL) {
    sprintf(err_msg, "Cannot open parm file %s.", parm_filename);
    errorMessage(err_msg);
  }
  
  nbond_mall= NMEM_MIN;
  nbond_now = 0;
  natm_tot  = maps->natm_tot;
  line_num  = 0;
  
  atom1 = new int[nbond_mall]-1;
  atom2 = new int[nbond_mall]-1;
  
  while((n=readOneEntry(parm_file,'~','[',']',entry_line, line_num))!=0) {
    if (n!=1) {
      sprintf(err_msg, "%s at line %d while reading file %s.",
	      (n==-1) ? "Syntax error" : "Internal error", 
	      line_num, parm_filename);
      errorMessage(err_msg);
    }
    n=getEntryName(entry_line, '~', '[', entry_name, def);
    if (n!=1) {
      sprintf(err_msg, "Syntax error at parm file line:\n%s", entry_line);
      errorMessage(err_msg);
    }
    if (strcasecmp(entry_name, "bond_def")==0) {
      nkey = 2;
      for ( i=1; i<=nkey; i++) { key_found[i] = 0; }
      while ((n=getKeyArgPair(def, key, arg, '\\', '{', '}', def))==1) {
	key_defined = 0;
	for ( i=1; i<=nkey; i++) {                  /*loop for key words.*/
	  if (strcasecmp(key, bond_def_keys[i])==0) {/*it's a keyword.  */
	    key_defined  = 1;
	    key_found[i] = 1;
	    switch (i) {
	    case 1:
	      sscanf(arg, "%d", &atom1_ind);
	      break;
	    case 2:
	      sscanf(arg, "%d", &atom2_ind);
	      break;
	    }/*switch */
	  }
	}/*end for*/
	if (key_defined==0) {
	  sprintf(err_msg, "Key word %s unknown in parmfile line:\n%s",
		  key, entry_line);
	  errorMessage(err_msg);
	}
      }/*end while: getKeyArgPair*/
      if (n!=0) {
	sprintf(err_msg, "Syntax error in parm file line:\n%s",entry_line);
	errorMessage(err_msg);
      }
      for ( i=1; i<=nkey; i++) { 
	if (key_found[i]==0) {
	  sprintf(err_msg, "Key word %s missed in parm file line:\n%s",
		  bond_def_keys[i], entry_line);
	  errorMessage(err_msg);
	}
      }
      if (atom1_ind<1 || atom1_ind>natm_tot || 
	  atom2_ind<1 || atom2_ind>natm_tot) {
	sprintf(err_msg, "BOND::setup(): Illegal atom index in parm file %s.",
		parm_filename);
	errorMessage(err_msg);
      }
      nbond_now++;
      if (nbond_now>nbond_mall) {
	nbond_mall += NMEM_MIN;
	atom1       = (int *) crealloc(atom1+1, nbond_mall*sizeof(int))-1;
	atom2       = (int *) crealloc(atom2+1, nbond_mall*sizeof(int))-1;
      }
      atom1[nbond_now] = atom1_ind;
      atom2[nbond_now] = atom2_ind;
    }/*end if: bond_def */
  }/*end while: readOneEntry*/
  fclose(parm_file);
  
  nbond = nbond_now;
  fk = new double[nbond]-1;
  r0 = new double[nbond]-1;
  
  for (ibond=1; ibond<=nbond; ibond++) {
    if (atom1[ibond]==atom2[ibond]) {
      sprintf(err_msg, "Ilegal bond #%d: same atom index.", ibond);
      errorMessage(err_msg);
    }
  }
  for (ibond=1; ibond<=nbond; ibond++) {
    atoms[1] = atom1[ibond];
    atoms[2] = atom2[ibond];
    exact_match_entry = NULL;
    wild_match_entry  = NULL;
    database_entry    = database;
    bond_check = 0;
    while (database_entry!=NULL && bond_check!=1) {
      bond_check = database_entry->match(maps, atoms, 1);
      if (bond_check==1) { exact_match_entry = database_entry; }
      if (bond_check==2) { wild_match_entry  = database_entry; }
      database_entry = database_entry->next;
    }/*end while */
    if (exact_match_entry==NULL && wild_match_entry==NULL) {
      sprintf(err_msg, "Bond is not found in database.\nindices %d %d \nlabels %s %s ",
	      atoms[1], atoms[2],
              maps->atm_label[atoms[1]],maps->atm_label[atoms[2]]);
      errorMessage(err_msg);
    }
    if (exact_match_entry!=NULL) { database_entry = exact_match_entry;}
    else                         { database_entry = wild_match_entry; }
    fk[ibond] = database_entry->fk;
    r0[ibond] = database_entry->r0;
  }/*end for: ibond */
  
  for (ibond=1; ibond<=nbond; ibond++) {
    for (jbond=ibond+1; jbond<=nbond; jbond++) {
      if (atom1[ibond]==atom1[jbond] &&
	  atom2[ibond]==atom2[jbond] ) {
	sprintf(err_msg, "Bonds %d and %d repeat.", ibond, jbond);
	errorMessage(err_msg);
      }
      if (atom1[ibond]==atom2[jbond] &&
	  atom2[ibond]==atom1[jbond] ) {
	sprintf(err_msg, "Bonds %d and %d repeat.", ibond, jbond);
	errorMessage(err_msg);
      }
    }
  }


  /*Set up the atom bond map. */
  nbond_atm = new int[natm_tot]-1;
  for (iatm=1; iatm<=natm_tot; iatm++) { nbond_atm[iatm] = 0; }
  for (ibond=1; ibond<=nbond; ibond++) {
    iatm = atom1[ibond];
    jatm = atom2[ibond];
    nbond_atm[iatm]++;
    nbond_atm[jatm]++;
  }
  nbond_atm_max = 0;
  for (iatm=1; iatm<=natm_tot; iatm++) {
    if (nbond_atm[iatm]>nbond_atm_max) { nbond_atm_max = nbond_atm[iatm];}
  }
  bond_atm_map = cmall_int_mat(1, natm_tot, 1, nbond_atm_max);
  for (iatm=1; iatm<=natm_tot; iatm++) {
    nbond_atm_tmp = 0;
    for (ibond=1; ibond<=nbond; ibond++) {
      if (iatm==atom1[ibond]) {
	bond_atm_map[iatm][++nbond_atm_tmp] = atom2[ibond];
      }
      if (iatm==atom2[ibond]) {
	bond_atm_map[iatm][++nbond_atm_tmp] = atom1[ibond];
      }
    }
    if (nbond_atm_tmp!=nbond_atm[iatm]) {
      sprintf(err_msg, "Internal error of setting up atom bond map.");
      errorMessage(err_msg);
    }
  }
  
#ifdef DEBUG      
  for (ibond=1; ibond<=nbond; ibond++) {
    printf("Bond %2d:  %d(%s)--%d(%s), {fk=%.8g, r0=%.8g}\n",ibond, 
   	   atom1[ibond], maps->atm_label[atom1[ibond]],
   	   atom2[ibond], maps->atm_label[atom2[ibond]],
  	   fk[ibond], r0[ibond]);
  }
#endif
  
  printf("Bonds setting up done.\n");
  printf("----------------------------------------------\n");


/*-------------------------------------------------------------------------*/
} /*end routine */
/*========================================================================*/



/*========================================================================*/
/*========================================================================*/
double Bond::energy(const Pos *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/
  int i, ibond;
  int atom1_ind, atom2_ind;

  double *x, *y, *z, *fx, *fy, *fz;
  double dx, dy, dz, r, rr, dr;
  double pre, vbond;
  
  x  = pos->x;
  y  = pos->y;
  z  = pos->z;
  fx = pos->fx;
  fy = pos->fy;
  fz = pos->fz;
  
  vbond = 0.0;

/*-------------------------------------------------------------------------*/
/* 0) Return if no bond term                                               */

  if (energy_term != 1) {
      return vbond;
  }/*endif*/

/*-------------------------------------------------------------------------*/
/* I) Evaluate energy                                                      */

  for (ibond=1; ibond<=nbond; ibond++) {
    atom1_ind = atom1[ibond];
    atom2_ind = atom2[ibond];
    dx = x[atom1_ind] - x[atom2_ind];
    dy = y[atom1_ind] - y[atom2_ind];
    dz = z[atom1_ind] - z[atom2_ind];
    rr = dx*dx+dy*dy+dz*dz;
    r  = sqrt(rr);
    dr = r-r0[ibond];
    vbond += 0.5*fk[ibond]*dr*dr;
    
    pre = -fk[ibond]*dr/r;
    fx[atom1_ind] += pre*dx;
    fy[atom1_ind] += pre*dy;
    fz[atom1_ind] += pre*dz;
    fx[atom2_ind] -= pre*dx;
    fy[atom2_ind] -= pre*dy;
    fz[atom2_ind] -= pre*dz;
  } /*end for*/
  return vbond;

/*-------------------------------------------------------------------------*/
} /*end routine */
/*========================================================================*/




/*========================================================================*/
/*========================================================================*/
double Bond::getPot(const Pos *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/
  int i, ibond;
  int atom1_ind, atom2_ind;

  double *x, *y, *z;
  double dx, dy, dz, r, rr, dr;
  double vbond;
  
  x  = pos->x;
  y  = pos->y;
  z  = pos->z;
  
  vbond = 0.0;

/*-------------------------------------------------------------------------*/
/* 0) Return if no bond term                                               */

  if (energy_term != 1) {
      return vbond;
  }/*endif*/

/*-------------------------------------------------------------------------*/
/* I) Evaluate energy                                                      */
  
  for (ibond=1; ibond<=nbond; ibond++) {
    atom1_ind = atom1[ibond];
    atom2_ind = atom2[ibond];
    dx = x[atom1_ind] - x[atom2_ind];
    dy = y[atom1_ind] - y[atom2_ind];
    dz = z[atom1_ind] - z[atom2_ind];
    rr = dx*dx+dy*dy+dz*dz;
    r  = sqrt(rr);
    dr = r-r0[ibond];
    vbond += 0.5*fk[ibond]*dr*dr;     
  } /*end for*/
  return vbond;
  
/*-------------------------------------------------------------------------*/
} /*end routine */
/*========================================================================*/
