#include "../include/standard.h"
#include "../include/typedefs.h"
#include "../include/definitions.h"
#include "../include/utillib.h"
#include "../include/cmalloc.h"

#define DEBUG_OFF

/*========================================================================*/
/*========================================================================*/
void Bond::setup_enm(const General *general, const Maps *maps,
                                            const Pos  *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/

  int natm_tot;

  int iatm, jatm, ibond_enm;
  int nbond_mall, nbond_now;

  double A_fk, B_fk, C_fk;
  double r0_fk;
  double enm_cutoff;
  double dx, dy, dz, r, rr, dr;
  char   err_msg[LINE_NCHARMAX];

  double  *x  = pos->x;
  double  *y  = pos->y;
  double  *z  = pos->z;
 
  printf("----------------------------------------------------\n");
  printf("Setting up bonds for the Elastic Network Model \n...\n");


/*-------------------------------------------------------------------------*/
/* I) Setup topology based on position and enm_cutoff                      */
 
  nbond_mall = NMEM_MIN;
  nbond_now  = 0;
  natm_tot   = maps->natm_tot;
  atom1_enm = new int[nbond_mall]-1;
  atom2_enm = new int[nbond_mall]-1;
  
  enm_cutoff = 10.0/BOHR; /* 10 Angstrom */
  for (iatm=1;iatm<=natm_tot-1;iatm++) {
    for (jatm=iatm+1;jatm<=natm_tot;jatm++) {
      dx = x[iatm]-x[jatm];
      dy = y[iatm]-y[jatm];
      dz = z[iatm]-z[jatm];
      rr = dx*dx + dy*dy + dz*dz;
      r  = sqrt(rr);
      if ( r <= enm_cutoff ) {
         nbond_now++;
         if (nbond_now>nbond_mall) {
 	     nbond_mall += NMEM_MIN;
             atom1_enm=(int *)crealloc(atom1_enm+1,nbond_mall*sizeof(int))-1;
             atom2_enm=(int *)crealloc(atom2_enm+1,nbond_mall*sizeof(int))-1;
         }
 	 atom1_enm[nbond_now] = iatm;
         atom2_enm[nbond_now] = jatm;
      }/*endif*/
    }/*endfor*/
  }/*endfor*/
  nbond_enm = nbond_now;

/*-------------------------------------------------------------------------*/
/* II) Setup force constatnts                                              */

  fk_enm    = new double[nbond_enm]-1;
  r0_enm    = new double[nbond_enm]-1;  

  for (ibond_enm=1; ibond_enm<=nbond_enm; ibond_enm++) {
    if (atom1_enm[ibond_enm]==atom2_enm[ibond_enm]) {
      sprintf(err_msg, "Ilegal bond #%d: same atom index.", ibond_enm);
      errorMessage(err_msg);
    }
  }

/*-------------------------------------------------------------------------*/
/* III) Initialize r0 and fk from coordinates                              */

  /* For fk see. Konrad Hinsen et al., Proteins 33 417-429 (1998) */

  r0_fk = 3.0/BOHR;    
  for (ibond_enm=1; ibond_enm<=nbond_enm; ibond_enm++) {
      dx = x[atom1_enm[ibond_enm]] - x[atom2_enm[ibond_enm]];
      dy = y[atom1_enm[ibond_enm]] - y[atom2_enm[ibond_enm]];
      dz = z[atom1_enm[ibond_enm]] - z[atom2_enm[ibond_enm]];
      rr = dx*dx+dy*dy+dz*dz;
      r  = sqrt(rr);
      r0_enm[ibond_enm] = r;
      fk_enm[ibond_enm] = exp(-rr/(r0_fk*r0_fk));
  }/*endfor*/

/*-------------------------------------------------------------------------*/
/* IV) DEBUG                                                               */

#ifdef DEBUG      
  for (ibond_enm=1; ibond_enm<=nbond_enm; ibond_enm++) {
    printf("Bond %2d:  %d(%s)--%d(%s), {fk=%.8g, r0=%.8g}\n",ibond_enm, 
   	   atom1_enm[ibond_enm], maps->atm_label[atom1_enm[ibond_enm]],
   	   atom2_enm[ibond_enm], maps->atm_label[atom2_enm[ibond_enm]],
  	   fk_enm[ibond_enm], r0_enm[ibond_enm]);
  }
#endif
  
  printf(" Elastic Network Model setting up done.       \n");
  printf("----------------------------------------------\n");

/*-------------------------------------------------------------------------*/
} /*end routine */
/*=========================================================================*/


/*========================================================================*/
/*========================================================================*/
double Bond::energy_enm(const Pos *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/
  int i, ibond_enm;
  int atom1_ind, atom2_ind;

  double *x, *y, *z, *fx, *fy, *fz;
  double dx, dy, dz, r, rr, dr;
  double pre, vbond_enm;
  
  x  = pos->x;
  y  = pos->y;
  z  = pos->z;
  fx = pos->fx;
  fy = pos->fy;
  fz = pos->fz;
  
  vbond_enm = 0.0;

  for (ibond_enm=1; ibond_enm<=nbond_enm; ibond_enm++) {
    atom1_ind = atom1_enm[ibond_enm];
    atom2_ind = atom2_enm[ibond_enm];
    dx = x[atom1_ind] - x[atom2_ind];
    dy = y[atom1_ind] - y[atom2_ind];
    dz = z[atom1_ind] - z[atom2_ind];
    rr = dx*dx+dy*dy+dz*dz;
    r  = sqrt(rr);
    dr = r-r0_enm[ibond_enm];
    vbond_enm += 0.5*fk_enm[ibond_enm]*dr*dr;
    
    pre = -fk_enm[ibond_enm]*dr/r;
    fx[atom1_ind] += pre*dx;
    fy[atom1_ind] += pre*dy;
    fz[atom1_ind] += pre*dz;
    fx[atom2_ind] -= pre*dx;
    fy[atom2_ind] -= pre*dy;
    fz[atom2_ind] -= pre*dz;
  } /*end for*/
  return vbond_enm;

/*-------------------------------------------------------------------------*/
} /*end routine */
/*========================================================================*/




/*========================================================================*/
/*========================================================================*/
double Bond::getPot_enm(const Pos *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/
  int i, ibond_enm;
  int atom1_ind, atom2_ind;

  double *x, *y, *z;
  double dx, dy, dz, r, rr, dr;
  double vbond_enm;
  
  x  = pos->x;
  y  = pos->y;
  z  = pos->z;
  
  vbond_enm = 0.0;
  
  for (ibond_enm=1; ibond_enm<=nbond_enm; ibond_enm++) {
    atom1_ind = atom1_enm[ibond_enm];
    atom2_ind = atom2_enm[ibond_enm];
    dx = x[atom1_ind] - x[atom2_ind];
    dy = y[atom1_ind] - y[atom2_ind];
    dz = z[atom1_ind] - z[atom2_ind];
    rr = dx*dx+dy*dy+dz*dz;
    r  = sqrt(rr);
    dr = r-r0_enm[ibond_enm];
    vbond_enm += 0.5*fk_enm[ibond_enm]*dr*dr;     
  } /*end for*/
  return vbond_enm;
  
/*-------------------------------------------------------------------------*/
} /*end routine */
/*========================================================================*/
