#include "../include/standard.h" 
#include "../include/typedefs.h"
#include "../include/definitions.h"
#include "../include/utillib.h"
#include "../include/cmalloc.h"

#define DEBUG_OFF

/*========================================================================*/
/*========================================================================*/
void Constraint::setup(const Constraint_data_entry *database, const Maps *maps, 
                                                              const Pos  *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/

  int i, iii, ibond, jbond, iatm, jatm, n;
  int natm_tot, nkey, key_defined, line_num;
  int nbond_now, nbond_mall, atom1_ind, atom2_ind;
  int bond_check;
  int nbond_atm_max, nbond_atm_tmp;
  double fk_value, r0_value;
  
  char   err_msg[LINE_NCHARMAX];

  printf("----------------------------------------------\n");
  printf("Setting up constraints\n...\n");
  
/*-------------------------------------------------------------------------*/
/* I) Setup topology for bond type constraints                             */

  nbond_mall= NMEM_MIN;
  nbond_now = 0;
  natm_tot  = maps->natm_tot;
  line_num  = 0;
  
  atom1 = new int[nbond_mall]-1;
  atom2 = new int[nbond_mall]-1;
  fk    = new double[nbond_mall]-1;
  r0    = new double[nbond_mall]-1;
 
  while (database!=NULL) {
      if ((database->atom1)<1 || (database->atom1)>natm_tot || 
    	  (database->atom2)<1 || (database->atom2)>natm_tot) {
	  sprintf(err_msg, "Constraint::setup(): Illegal atom index in database file");
	  errorMessage(err_msg);
      }
      nbond_now++;
      if (nbond_now>nbond_mall) {
          nbond_mall += NMEM_MIN;
	  atom1       = (int *) crealloc(atom1+1, nbond_mall*sizeof(int))-1;
	  atom2       = (int *) crealloc(atom2+1, nbond_mall*sizeof(int))-1;
	  fk          = (double *) crealloc(atom1+1, nbond_mall*sizeof(double))-1;
	  r0          = (double *) crealloc(atom2+1, nbond_mall*sizeof(double))-1;
      }
      atom1[nbond_now] = database->atom1;
      atom2[nbond_now] = database->atom2;
      fk[nbond_now]    = database->fk;
      r0[nbond_now]    = database->r0; 
      database = database->next; 
  }/*end while database_entry*/ 
  
/*-------------------------------------------------------------------------*/
/* II) Checks on bond type constraints                                     */

  nbond = nbond_now;
  
  for (ibond=1; ibond<=nbond; ibond++) {
    if (atom1[ibond]==atom2[ibond]) {
      sprintf(err_msg, "Ilegal bond #%d: same atom index.", ibond);
      errorMessage(err_msg);
    }
  }
  
  for (ibond=1; ibond<=nbond; ibond++) {
    for (jbond=ibond+1; jbond<=nbond; jbond++) {
      if (atom1[ibond]==atom1[jbond] &&
	  atom2[ibond]==atom2[jbond] ) {
	sprintf(err_msg, "Bond constraints %d and %d repeat.", ibond, jbond);
	errorMessage(err_msg);
      }
      if (atom1[ibond]==atom2[jbond] &&
	  atom2[ibond]==atom1[jbond] ) {
	sprintf(err_msg, "Bond constraints %d and %d repeat.", ibond, jbond);
	errorMessage(err_msg);
      }
    }
  }

/*-------------------------------------------------------------------------*/
/* III) Setup the atom constraint bond map                                 */

  /* Set up the atom bond map. */
  nbond_atm = new int[natm_tot]-1;
  for (iatm=1; iatm<=natm_tot; iatm++) { nbond_atm[iatm] = 0; }
  for (ibond=1; ibond<=nbond; ibond++) {
    iatm = atom1[ibond];
    jatm = atom2[ibond];
    nbond_atm[iatm]++;
    nbond_atm[jatm]++;
  }
  nbond_atm_max = 0;
  for (iatm=1; iatm<=natm_tot; iatm++) {
    if (nbond_atm[iatm]>nbond_atm_max) { nbond_atm_max = nbond_atm[iatm];}
  }
  bond_atm_map = cmall_int_mat(1, natm_tot, 1, nbond_atm_max);
  for (iatm=1; iatm<=natm_tot; iatm++) {
    nbond_atm_tmp = 0;
    for (ibond=1; ibond<=nbond; ibond++) {
      if (iatm==atom1[ibond]) {
	bond_atm_map[iatm][++nbond_atm_tmp] = atom2[ibond];
      }
      if (iatm==atom2[ibond]) {
	bond_atm_map[iatm][++nbond_atm_tmp] = atom1[ibond];
      }
    }
    if (nbond_atm_tmp!=nbond_atm[iatm]) {
      sprintf(err_msg, "Internal error of setting up atom bond map.");
      errorMessage(err_msg);
    }
  }
  
/*-------------------------------------------------------------------------*/
/* IV) DEBUG                                                               */

#ifdef DEBUG      
  for (ibond=1; ibond<=nbond; ibond++) {
    printf("Bond %2d:  %d(%s)--%d(%s), {fk=%.8g, r0=%.8g}\n",ibond, 
   	   atom1[ibond], maps->atm_label[atom1[ibond]],
   	   atom2[ibond], maps->atm_label[atom2[ibond]],
  	   fk[ibond], r0[ibond]);
  }
  getchar();
#endif
  
  printf("Constraints setting up done.\n");
  printf("----------------------------------------------\n");

/*-------------------------------------------------------------------------*/
} /*end routine */
/*========================================================================*/



/*========================================================================*/
/*========================================================================*/
double Constraint::energy(const Pos *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/
  int i, ibond;
  int atom1_ind, atom2_ind;

  double *x, *y, *z, *fx, *fy, *fz;
  double dx, dy, dz, r, rr, dr;
  double pre, vbond;
  
  x  = pos->x;
  y  = pos->y;
  z  = pos->z;
  fx = pos->fx;
  fy = pos->fy;
  fz = pos->fz;
  
  vbond = 0.0;

/*-------------------------------------------------------------------------*/
/* 0) Return if no bond term                                               */

  if (energy_term != 1) {
      return vbond;
  }/*endif*/

/*-------------------------------------------------------------------------*/
/* I) Evaluate energy                                                      */

  for (ibond=1; ibond<=nbond; ibond++) {
    atom1_ind = atom1[ibond];
    atom2_ind = atom2[ibond];
    dx = x[atom1_ind] - x[atom2_ind];
    dy = y[atom1_ind] - y[atom2_ind];
    dz = z[atom1_ind] - z[atom2_ind];
    rr = dx*dx+dy*dy+dz*dz;
    r  = sqrt(rr);
    dr = r-r0[ibond];
    vbond += 0.5*fk[ibond]*dr*dr;
    
    pre = -fk[ibond]*dr/r;
    fx[atom1_ind] += pre*dx;
    fy[atom1_ind] += pre*dy;
    fz[atom1_ind] += pre*dz;
    fx[atom2_ind] -= pre*dx;
    fy[atom2_ind] -= pre*dy;
    fz[atom2_ind] -= pre*dz;
  } /*end for*/
  return vbond;

/*-------------------------------------------------------------------------*/
} /*end routine */
/*========================================================================*/




/*========================================================================*/
/*========================================================================*/
double Constraint::getPot(const Pos *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/
  int i, ibond;
  int atom1_ind, atom2_ind;

  double *x, *y, *z;
  double dx, dy, dz, r, rr, dr;
  double vbond;
  
  x  = pos->x;
  y  = pos->y;
  z  = pos->z;
  
  vbond = 0.0;

/*-------------------------------------------------------------------------*/
/* 0) Return if no bond term                                               */

  if (energy_term != 1) {
      return vbond;
  }/*endif*/

/*-------------------------------------------------------------------------*/
/* I) Evaluate energy                                                      */
  
  for (ibond=1; ibond<=nbond; ibond++) {
    atom1_ind = atom1[ibond];
    atom2_ind = atom2[ibond];
    dx = x[atom1_ind] - x[atom2_ind];
    dy = y[atom1_ind] - y[atom2_ind];
    dz = z[atom1_ind] - z[atom2_ind];
    rr = dx*dx+dy*dy+dz*dz;
    r  = sqrt(rr);
    dr = r-r0[ibond];
    vbond += 0.5*fk[ibond]*dr*dr;     
  } /*end for*/
  return vbond;
  
/*-------------------------------------------------------------------------*/
} /*end routine */
/*========================================================================*/
