#define DEBUG_OFF
#define CRYO_EM_OFF

#include "../include/standard.h"
#include "../include/typedefs.h"
#include "../include/definitions.h"
#include "../include/utillib.h"
#include "../include/cmalloc.h"

#ifdef CRYO_EM
#include "../include/cryo_em/Sim_Score_Matrix.h"
#endif

/*=========================================================================*/
/*=========================================================================*/
void Cryo_em::setup(const Pot_data_entry *database, const System *system)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/

  int iorient, nentry;
  int iatm, icalpha, icount;

/*-------------------------------------------------------------------------*/
/* I) Count the number of entries in database                              */

  const Pot_data_entry *database_entry;

  nentry = 0;
  database_entry    = database;
  while (database_entry!=NULL) {
    nentry++;
    database_entry = database_entry->next;
  }/*end while */

/*-------------------------------------------------------------------------*/
/* II) Allocate variables                                                  */

  norient = nentry;
  pixel_size   = new double[norient]-1;
  resol_blur   = new double[norient]-1;
  ea_az        = new double[norient]-1;
  ea_alt       = new double[norient]-1;
  ea_phi       = new double[norient]-1;
  ea_range     = new double[norient]-1;
  ea_interval  = new double[norient]-1;
  energy_scale = new double[norient]-1;
  expermnt_filename = (char **)cmall_int_mat(1,norient,0,NCHARMAX);
#ifdef CRYO_EM
  eman2_pa          = new PointArray();
#endif

/*-------------------------------------------------------------------------*/
/* III) Assign variables                                                   */

  nentry = 0;
  database_entry    = database;
  while (database_entry!=NULL) {
    nentry++;
    pixel_size[nentry] = database_entry->pixel_size;
    resol_blur[nentry] = database_entry->resol_blur;
    ea_az[nentry] = database_entry->ea_az;
    ea_alt[nentry] = database_entry->ea_alt;
    ea_phi[nentry] = database_entry->ea_phi;
    ea_range[nentry] = database_entry->ea_range;
    ea_interval[nentry] = database_entry->ea_interval;
    energy_scale[nentry] = database_entry->energy_scale;
    strcpy(expermnt_filename[nentry],database->expermnt_filename);
    database_entry = database_entry->next;
  }/*end while */

/*-------------------------------------------------------------------------*/
/* IV) Define calpha trace                                                 */
/*     In case of nucleic acids CA trace is filled with C4'                */

  icount = 0;
  for (iatm=1;iatm<=system->pos.natom;iatm++) {
    if (strcasecmp(system->maps.upac_atm_label[iatm],"CA")==0){icount++;}
    if (strcasecmp(system->maps.upac_atm_label[iatm],"C4'")==0){icount++;}
  }/*endfor*/
  ncalpha = icount;
  
  ind_calpha = new int[ncalpha] - 1;
  icount = 0;
  for (iatm=1;iatm<=system->pos.natom;iatm++) {
    if (strcasecmp(system->maps.upac_atm_label[iatm],"CA")==0){
        icount++;
        ind_calpha[icount] = iatm;
    }/*endif*/
    if (strcasecmp(system->maps.upac_atm_label[iatm],"C4'")==0){
        icount++;
        ind_calpha[icount] = iatm;
    }/*endif*/
  }/*endfor*/

#ifdef CRYO_EM
/*-------------------------------------------------------------------------*/
/* V) Set size of point array                                              */

  eman2_pa->set_number_points(ncalpha);

/*-------------------------------------------------------------------------*/
/* VI) Initialize Sim_Score_Matrix                                         */

  for (iorient=1;iorient<=norient;iorient++) {
       ssm.set_sim_score_matrix_dim(ea_range[iorient],ea_interval[iorient]);
       ssm.init_sim_score_matrix();
  }/*endfor*/

#endif

/*-------------------------------------------------------------------------*/
/* VII) DEBUG                                                              */

#ifdef DEBUG
  printf("norient %d\n",norient);
  for (iorient=1;iorient<=norient;iorient++) {
    printf("i %d: pixel_size %g, resol_blur %g\n",iorient,
                                                  pixel_size[iorient],
                                                  resol_blur[iorient]);
    printf("i %d: ea_az %g, ea_alt %g, ea_phi %g\n",iorient,
                                                    ea_az[iorient],
                                                    ea_alt[iorient],
                                                    ea_phi[iorient]);
    printf("i %d: ea_range %g, ea_interval %g\n",iorient,
                                                 ea_range[iorient],
                                                 ea_interval[iorient]);
    printf("i %d: expermnt_file %s\n",iorient,expermnt_filename[iorient]);
    getchar();
  }/*endfor*/

  for (icalpha=1;icalpha<=ncalpha;icalpha++) {
       iatm = ind_calpha[icalpha];
       printf("icalpha %d is %d. %s\n",icalpha,
                                       iatm,
                                       system->maps.upac_atm_label[iatm]);
  }/*endfor*/
  getchar();
#endif


/*-------------------------------------------------------------------------*/
} /*end routine */
/*=========================================================================*/


/*=========================================================================*/
/*=========================================================================*/
double Cryo_em::energy(const Pos *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/

  int i, iorient;
  double vcryo_em, scale;
  double *x, *y, *z;
  
  x  = pos->x;
  y  = pos->y;
  z  = pos->z;
  double *points;

  vcryo_em = 0.0;

#ifdef CRYO_EM
/*-------------------------------------------------------------------------*/
/* I) Return if no energy_term                                             */


  if (energy_term != 1) {return vcryo_em;}


/*-------------------------------------------------------------------------*/
/* II) Set CA positions in eman2_ca                                        */

  points = eman2_pa->get_points_array();

  for (i=0;i<ncalpha;i++) {
    points[i*4]   = x[ind_calpha[i+1]]*BOHR; 
    points[i*4+1] = y[ind_calpha[i+1]]*BOHR;
    points[i*4+2] = z[ind_calpha[i+1]]*BOHR;
    points[i*4+3] = 6;
  }/*endfor*/
  
/*-------------------------------------------------------------------------*/
/* III) Accumulate orientation score                                       */

  for (iorient=1;iorient<=norient;iorient++) {
	
// calculate the original similarity score matrix

#ifdef DEBUG
  printf("Before cryo_em.energy Check Point IV\n");
     getchar();
#endif



    ssm.calc_sim_score_matrix(eman2_pa,expermnt_filename[iorient],ea_az[iorient],ea_alt[iorient],
       		              ea_phi[iorient],pixel_size[iorient],resol_blur[iorient]);   


#ifdef DEBUG
  printf("Before cryo_em.energy Check Point V\n");
     getchar();
#endif

	
     //    ssm.dump_sim_score_matrix();
    ssm.order_sim_score_matrix();
     // ssm.dump_sim_score_matrix();
 
    scale     = energy_scale[iorient];
    vcryo_em += scale*ssm.get_sim_score_matrix()[0].get_score();

  }/*endfor*/


#endif /*CRYO_EM*/

  return vcryo_em;

/*-------------------------------------------------------------------------*/
} /*end routine */
/*========================================================================*/

/*========================================================================*/
/*========================================================================*/
double Cryo_em::getPot(const Pos *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/

  return energy(pos);
  
/*-------------------------------------------------------------------------*/
} /*end routine */
/*========================================================================*/
