#include "../include/standard.h"
#include "../include/typedefs.h"
#include "../include/definitions.h"
#include "../include/utillib.h"
#include "../include/cmalloc.h"

#define DEBUG
#undef DEBUG

/*=========================================================================*/
/*-------------------------------------------------------------------------*/
void Residue_bond::setup_1pt(Maps* maps, Residue* residue, int** bond_res)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/

  int ires1, ires2, ires, iresbond; 
  int ichain, nchain, bondmatch, length;
  int ibond, ibend, itors, ionfo, iexclude;
  int iconn, nconn;
  int **head_tail;

  int inter_on_3bond_conn = maps->inter_on_3bond_conn;
  int inter_on_4bond_conn = maps->inter_on_4bond_conn;

  char   err_msg[LINE_NCHARMAX];  

  int nres     = maps->nres;
  int nresbond = maps->nresbond; 

/*-------------------------------------------------------------------------*/
/* 0) Identify the chains to bond                                          */
/*    Assume that only residues with neigboring indices might bond         */

  head_tail = cmall_int_mat(1, nres, 1, 2);

  nchain = 1;
  head_tail[nchain][1] = 1;
  for (ires=1;ires<=nres-1;ires++) {
     bondmatch = 0;
     for (iresbond=1;iresbond<=nresbond;iresbond++) {
          ires1 = bond_res[iresbond][1];
          ires2 = bond_res[iresbond][2];
          if ( ((ires==ires1) && (ires+1==ires2)) ||
               ((ires+1==ires1) && (ires==ires2))  ) {
	       bondmatch = 1;
          }
     }/*end for*/
     if (bondmatch==0) {
         head_tail[nchain][2] = ires;
         nchain++;
         head_tail[nchain][1] = ires+1;         
     }
  }/*end for*/
  head_tail[nchain][2] = nres; 

/*-------------------------------------------------------------------------*/
/* I) Count the number of bonds,bends,tors,impr,onfo,exclude               */
/*    Only linear chains are involved                                      */
/*    nexclude = 3*length, since at maximum each atom does not interact    */
/*               with 3 atoms having larger index                          */
 
  nbond = 0;
  nbend = 0;
  ntors = 0;
  nimpr = 0;
  nonfo = 0;
  nexclude = 0; 

  for (ichain=1;ichain<=nchain;ichain++) {
       length = head_tail[ichain][2]-head_tail[ichain][1]+1;
       nbond += MAX(0,length-1);
       nbend += MAX(0,length-2);
       ntors += MAX(0,length-3);
       nonfo += MAX(0,length-3);
       if (inter_on_3bond_conn==1) {nexclude += MAX(0,length-1);}
       else if (inter_on_4bond_conn==1) {nexclude += MAX(0,2*(length-2)+1);}
       else {nexclude += ((length==1)?(0):(MAX(1,3*(length-3)+3)));}
  }/*end for*/
  
/*-------------------------------------------------------------------------*/
/* II) Allocate variables                                                  */

  bond_atom     =  cmall_int_mat(1, nbond, 1, 2);
  bend_atom     =  cmall_int_mat(1, nbend, 1, 3);
  tors_atom     =  cmall_int_mat(1, ntors, 1, 4);
  impr_atom     =  cmall_int_mat(1, nimpr, 1, 4);
  onfo_atom     =  cmall_int_mat(1, nonfo, 1, 2);
  exclude_atom  =  cmall_int_mat(1, nexclude, 1, 2);

  bond    =  new int[nbond] - 1;
  bend    =  new int[nbend] - 1;
  tors    =  new int[ntors] - 1;
  impr    =  new int[nimpr] - 1;
  onfo    =  new int[nonfo] - 1;
  exclude =  new int[nexclude] - 1;
    
/*-------------------------------------------------------------------------*/
/* III) Fill in the variables                                              */
/*     1pt : residue[].branch_root_f[0][0] or residue[].branch_root_s[0][0]*/ 

   ibond = 0;
   ibend = 0;
   itors = 0;
   ionfo = 0;
   iexclude = 0;
   for (ichain=1;ichain<=nchain;ichain++) {       
    for (ires=head_tail[ichain][1];ires<=head_tail[ichain][2]-1;ires++) {
	 ibond++;
         if (ibond>nbond) {
            sprintf(err_msg,"ibond: %d > nbond: %d in Residue_bond::setup_1pt",
                                                                  ibond,nbond);
            errorMessage(err_msg);
         }/*end if*/
         bond_atom[ibond][1] = residue[ires].branch_root_f[0][0];
         bond_atom[ibond][2] = residue[ires+1].branch_root_f[0][0];
    }/*end for*/
    for (ires=head_tail[ichain][1];ires<=head_tail[ichain][2]-2;ires++) {
         ibend++;
         if (ibend>nbend) {
           sprintf(err_msg,"ibend: %d > nbend: %d in Residue_bond::setup_1pt",
      	                                                         ibend,nbend);
           errorMessage(err_msg);
         }/*end if*/
         bend_atom[ibend][1] = residue[ires].branch_root_f[0][0];
         bend_atom[ibend][2] = residue[ires+1].branch_root_f[0][0];
         bend_atom[ibend][3] = residue[ires+2].branch_root_f[0][0];
    }/*end for*/
    for (ires=head_tail[ichain][1];ires<=head_tail[ichain][2]-3;ires++) {
	 itors++;
         if (itors>ntors) {
           sprintf(err_msg,"itors: %d > ntors: %d in Residue_bond::setup_1pt",
	                                                         itors,ntors);
           errorMessage(err_msg);
         }/*end if*/
	 tors_atom[itors][1] = residue[ires].branch_root_f[0][0];
         tors_atom[itors][2] = residue[ires+1].branch_root_f[0][0];
	 tors_atom[itors][3] = residue[ires+2].branch_root_f[0][0];
         tors_atom[itors][4] = residue[ires+3].branch_root_f[0][0];
         ionfo++;
         if (ionfo>nonfo) {
           sprintf(err_msg,"ionfo: %d > nonfo: %d in Residue_bond::setup_1pt",
		                                                 ionfo,nonfo);
           errorMessage(err_msg);
         }/*end if*/
	 onfo_atom[ionfo][1] = residue[ires].branch_root_f[0][0];
         onfo_atom[ionfo][2] = residue[ires+3].branch_root_f[0][0];
    }/*end for*/
    for (ires=head_tail[ichain][1];ires<=head_tail[ichain][2];ires++) {
      if      (inter_on_3bond_conn==1) {nconn=1;}
      else if (inter_on_4bond_conn==1) {nconn=2;}
      else                             {nconn=3;}  
      for (iconn=1;iconn<=nconn;iconn++) {
       if (ires+iconn<=head_tail[ichain][2]) {
        iexclude++;
        if (iexclude>nexclude) {
           sprintf(err_msg,"iexclude: %d > nexclude: %d in setup_1pt",
		                                   iexclude,nexclude);
           errorMessage(err_msg);
        }/*end if*/
        exclude_atom[iexclude][1] = residue[ires].branch_root_f[0][0];
        exclude_atom[iexclude][2] = residue[ires+iconn].branch_root_f[0][0];
       }/*end if*/
      }/*end for*/
    }/*end for*/
   }/*end for*/
   cfree_int_mat(head_tail,1,nres,1,2);
    
/*-------------------------------------------------------------------------*/
/* IV) Check predicted length vs. counted length                           */
   if (ibond!=nbond) { 
       sprintf(err_msg,"ibond: %d != nbond: %d in setup_1pt",
                                          iexclude,nexclude);
       errorMessage(err_msg);
   }/*end if*/
   if (ibend!=nbend) {
       sprintf(err_msg,"ibend: %d != nbend: %d in Residue_bond::setup_1pt",
      	                                                    ibend,nbend);
       errorMessage(err_msg);
   }/*end if*/
   if (itors!=ntors) {
       sprintf(err_msg,"itors: %d != ntors: %d in Residue_bond::setup_1pt",
                                                             itors,ntors);
       errorMessage(err_msg);
   }/*end if*/
   if (ionfo!=nonfo) {
       sprintf(err_msg,"ionfo: %d != nonfo: %d in Residue_bond::setup_1pt",
	                                                     ionfo,nonfo);
       errorMessage(err_msg);
   }/*end if*/
   if (iexclude!=nexclude) {
       sprintf(err_msg,"iexclude: %d != nexclude: %d in setup_1pt",
	                                       iexclude,nexclude);
       errorMessage(err_msg);
   }/*end if*/
/*-------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*/
} /*end routine */
/*-------------------------------------------------------------------------*/
/*=========================================================================*/


/*=========================================================================*/
/*-------------------------------------------------------------------------*/
void Residue_bond::setup(Maps* maps, Residue* res1, Residue* res2, Pos *pos)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/

  int iatm, jatm;
  int ibranch1, ibranch2;
  int iibranch1, iibranch2;

  int valid;
  int ibond, ibend, itors, iimpr, ionfo;
  int inter_on_3bond_conn = maps->inter_on_3bond_conn;
  int inter_on_4bond_conn = maps->inter_on_4bond_conn;
  int nmem_min_exclude;

  int size, iatm_next, iatm_prev, bonding;
  char   err_msg[LINE_NCHARMAX];

/*-------------------------------------------------------------------------*/
/* I) Allocate variables                                                   */
  
  nmem_min_exclude = 20*NMEM_MIN;

  bond_atom     =  cmall_int_mat(1, NMEM_MIN, 1, 2);
  bend_atom     =  cmall_int_mat(1, NMEM_MIN, 1, 3);
  tors_atom     =  cmall_int_mat(1, NMEM_MIN, 1, 4);
  impr_atom     =  cmall_int_mat(1, NMEM_MIN, 1, 4);
  onfo_atom     =  cmall_int_mat(1, NMEM_MIN, 1, 2);
  exclude_atom  =  cmall_int_mat(1, nmem_min_exclude, 1, 2);

/*-------------------------------------------------------------------------*/
/* II) Fill in variables                                                   */

  strcpy(residue1_name,res1->residue_name);     
  strcpy(residue2_name,res2->residue_name);
  residue1_index = res1->residue_index;                
  residue2_index = res2->residue_index;                

#define DEBUG
#undef DEBUG
#ifdef DEBUG
  printf("Residue_bond::setup\n"); 
  printf("residue1: %s\n",residue1_name);
  printf("residue2: %s\n",residue2_name);
#endif
#undef DEBUG

  nbond = 1;
  bond_atom[nbond][1] = res1->branch_root_s[0][0];
  bond_atom[nbond][2] = res2->branch_root_f[0][0];

  nbend = 0;
  for (ibranch1=1;ibranch1<=res1->nbranch_s1;ibranch1++) {
      nbend++;
      if (nbend>NMEM_MIN) {
	sprintf(err_msg,"Allocate more bend in Residue_bond:setup");
	errorMessage(err_msg);
      }
      bend_atom[nbend][1] =  res2->branch_root_f[0][0];
      bend_atom[nbend][2] =  res1->branch_root_s[0][0];
      bend_atom[nbend][3] =  res1->branch_root_s[ibranch1][0];
  }/*end for*/
  for (ibranch2=1;ibranch2<=res2->nbranch_f1;ibranch2++) {
      nbend++;
      if (nbend>NMEM_MIN) {
	sprintf(err_msg,"Allocate more bend in Residue_bond:setup");
	errorMessage(err_msg);
      }
      bend_atom[nbend][1] =  res1->branch_root_s[0][0];
      bend_atom[nbend][2] =  res2->branch_root_f[0][0];
      bend_atom[nbend][3] =  res2->branch_root_f[ibranch2][0];
  }/*end for*/

  ntors = 0;
  for (ibranch1=1;ibranch1<=res1->nbranch_s1;ibranch1++) {
      for (ibranch2=1;ibranch2<=res2->nbranch_f1;ibranch2++) {
          ntors++;
          if (ntors>NMEM_MIN) {
	     sprintf(err_msg,"Allocate more tors in Residue_bond:setup");
	     errorMessage(err_msg);
          }
          tors_atom[ntors][1] =  res1->branch_root_s[ibranch1][0];
          tors_atom[ntors][2] =  res1->branch_root_s[0][0];
          tors_atom[ntors][3] =  res2->branch_root_f[0][0];
          tors_atom[ntors][4] =  res2->branch_root_f[ibranch2][0];   
      }/*end for*/
  }/*end for*/
  for (ibranch1=1;ibranch1<=res1->nbranch_s1;ibranch1++) {
      for (iibranch1=1;iibranch1<=res1->nbranch_s2[ibranch1];iibranch1++) {
          ntors++;
          if (ntors>NMEM_MIN) {
	     sprintf(err_msg,"Allocate more tors in Residue_bond:setup");
	     errorMessage(err_msg);
          }
          tors_atom[ntors][1] = res2->branch_root_f[0][0];
          tors_atom[ntors][2] = res1->branch_root_s[0][0]; 
          tors_atom[ntors][3] = res1->branch_root_s[ibranch1][0];
          tors_atom[ntors][4] = res1->branch_root_s[ibranch1][iibranch1];
      }/*end for*/
  }/*end for*/
  for (ibranch2=1;ibranch2<=res2->nbranch_f1;ibranch2++) {
      for (iibranch2=1;iibranch2<=res2->nbranch_f2[ibranch2];iibranch2++) {
          ntors++;
          if (ntors>NMEM_MIN) {
	     sprintf(err_msg,"Allocate more tors in Residue_bond:setup");
	     errorMessage(err_msg);
          }
          tors_atom[ntors][1] = res1->branch_root_s[0][0]; 
          tors_atom[ntors][2] = res2->branch_root_f[0][0];
          tors_atom[ntors][3] = res2->branch_root_f[ibranch2][0];
          tors_atom[ntors][4] = res2->branch_root_f[ibranch2][iibranch2];
      }/*end for*/
  }/*end for*/

  nonfo = 0;
  for (itors=1;itors<=ntors;itors++) {
      nonfo++;
      if (nonfo>NMEM_MIN) {
	 sprintf(err_msg,"Allocate more onfo in Residue_bond:setup");
	 errorMessage(err_msg);
      }
      onfo_atom[nonfo][1] =  tors_atom[itors][1];
      onfo_atom[nonfo][2] =  tors_atom[itors][4];
  }/*end for*/

  nexclude = 0;
  for (iatm=1;iatm<=res1->natom;iatm++) {
      for (jatm=1;jatm<=res2->natom;jatm++) {
	   valid = 1;
           if (maps->cancel_res_inter_neighbor==1) {/* if inter off btw   */
	     valid=0;                               /* atoms in neigh. res*/
           }/*end if*/                        
           for (ibond=1; ibond<=nbond; ibond++) {
	        if ((res1->atom[iatm]==bond_atom[ibond][1] ||
	             res1->atom[iatm]==bond_atom[ibond][2] ) &&
	            (res2->atom[jatm]==bond_atom[ibond][1] ||
	             res2->atom[jatm]==bond_atom[ibond][2] )  ) { valid = 0;}
           }/*end for*/
           if ((valid==1)&&(inter_on_3bond_conn==0)) {
              for (ibend=1; ibend<=nbend; ibend++) {
	        if ((res1->atom[iatm]==bend_atom[ibend][1] ||
	             res1->atom[iatm]==bend_atom[ibend][2] ||
	             res1->atom[iatm]==bend_atom[ibend][3] ) &&
	            (res2->atom[jatm]==bend_atom[ibend][1] ||
	             res2->atom[jatm]==bend_atom[ibend][2] ||
	             res2->atom[jatm]==bend_atom[ibend][3] )  ) { valid = 0;}
              }/*end for*/
           }/*end if*/
           if ((valid==1)&&(inter_on_3bond_conn==0)&&(inter_on_4bond_conn==0)) {
              for (ionfo=1; ionfo<=nonfo; ionfo++) {
	        if ((res1->atom[iatm]==onfo_atom[ionfo][1] ||
	             res1->atom[iatm]==onfo_atom[ionfo][2] ) &&
                    (res2->atom[jatm]==onfo_atom[ionfo][1] ||
	             res2->atom[jatm]==onfo_atom[ionfo][2] )  ) { valid = 0;}
              }/*end for*/
              for (itors=1; itors<=ntors; itors++) {
	        if ((res1->atom[iatm]==tors_atom[itors][1] ||
	             res1->atom[iatm]==tors_atom[itors][2] ||
	             res1->atom[iatm]==tors_atom[itors][3] ||
	             res1->atom[iatm]==tors_atom[itors][4] ) &&
	            (res2->atom[jatm]==tors_atom[itors][1] ||
	             res2->atom[jatm]==tors_atom[itors][2] ||
	             res2->atom[jatm]==tors_atom[itors][3] ||
	             res2->atom[jatm]==tors_atom[itors][4] )  ) { valid = 0;}
              }/*end for*/
           }/*end if*/
           if (valid==0) {
              nexclude++;
              if (nexclude>nmem_min_exclude) {
		sprintf(err_msg,"Allocate more exclude in Residue_bond:setup");
	        errorMessage(err_msg);
              }
              exclude_atom[nexclude][1] = res1->atom[iatm];
              exclude_atom[nexclude][2] = res2->atom[jatm];
           }/*end if*/
      }/*end for*/
  }/*end for*/

/*-------------------------------------------------------------------------*/
/* III) Define improper torsions                                           */
/*      convention used 0 : root atom in other residue                     */
/*      res1_tail(bimpr_atom_s) (=) res2_head(bimpr_atom_f)                */
/*      planed for nucleic acids and proteins                              */
/*      cannot be applied for PRES terminal groups                         */

  size = 4;
  nimpr = 0;
  if (res1->nbimpr==2) {/* res1 has 2nd bimpr */
     nimpr++;
     if (nimpr>NMEM_MIN) {
	sprintf(err_msg,"Allocate more impr in Residue_bond:setup %s");
	errorMessage(err_msg);
     }
     for (iatm=1;iatm<=size;iatm++) {
         if (res1->bimpr_atom_s[iatm]==0) {/* assume 1st root in res2 */
             impr_atom[nimpr][iatm] = res2->branch_root_f[0][0];
         }
         else {
             impr_atom[nimpr][iatm] = res1->bimpr_atom_s[iatm];  
         }/* end if */
     }/* end for */
  }/* end if */

  if (res2->nbimpr!=0) {/* res2 has 1st bimpr */
     nimpr++;
     if (nimpr>NMEM_MIN) {
	sprintf(err_msg,"Allocate more impr in Residue_bond:setup %s");
	errorMessage(err_msg);
     }/* end if */
     for (iatm=1;iatm<=size;iatm++) {
         if (res2->bimpr_atom_f[iatm]==0) {/* assume 1st root in res2 */
             impr_atom[nimpr][iatm] = res1->branch_root_s[0][0];
         }
         else {
             impr_atom[nimpr][iatm] = res2->bimpr_atom_f[iatm];  
         }/* end if */
     }/* end for */
  }/* end if */

/*-------------------------------------------------------------------------*/
/* IV) Allocate global index variables                                     */

  bond    =  new int[nbond] - 1;
  bend    =  new int[nbend] - 1;
  tors    =  new int[ntors] - 1;
  impr    =  new int[nimpr] - 1;
  onfo    =  new int[nonfo] - 1;
  exclude =  new int[nexclude] - 1;

/*-------------------------------------------------------------------------*/
} /*end routine */
/*-------------------------------------------------------------------------*/
/*=========================================================================*/


/*=========================================================================*/
/*-------------------------------------------------------------------------*/
void Residue_bond::write_alldata(Maps *maps)
/*-------------------------------------------------------------------------*/
{ /*begin routine */
/*-------------------------------------------------------------------------*/

  int iatm;
  int ibond, ibend, itors, iimpr, ionfo, iexclude;
  int ibranch, iibranch;

  printf("~~~~~~~~~~~~~~~~~~~~~~ Residue_bond ~~~~~~~~~~~~~~~~~~~~~~\n");
  printf("residue1_name : %s\n",residue1_name);
  printf("residue2_name : %s\n",residue2_name);


  printf("----------bonds----------\n");
  printf("nbond %d\n",nbond);
  printf("ind_glob ind_loc-> ");
  printf(" ind_atm1 : ind_atm2 |");
  printf(" lbl_atm1 : lbl_atm2\n");
  for (ibond=1;ibond<=nbond;ibond++) {
      printf("%d %d-> ",bond[ibond],ibond);
      printf("%d : %d | ",bond_atom[ibond][1],bond_atom[ibond][2]);
      printf("%s : %s\n",maps->atm_label[bond_atom[ibond][1]], 
                         maps->atm_label[bond_atom[ibond][2]]);
  }/*end for*/

  printf("----------bends----------\n");
  printf("nbend %d\n",nbend);
  printf("ind_glob ind_loc-> ");
  printf(" iatm1 : iatm2 : iatm3 |");
  printf(" latm1 : latm2 : latm3\n");
  for (ibend=1;ibend<=nbend;ibend++) {
      printf("%d %d-> ",bend[ibend],ibend);
      printf("%d : %d : %d | ",bend_atom[ibend][1],bend_atom[ibend][2],
                                                   bend_atom[ibend][3]);
      printf("%s : %s : %s\n",maps->atm_label[bend_atom[ibend][1]], 
                              maps->atm_label[bend_atom[ibend][2]],
                              maps->atm_label[bend_atom[ibend][3]]);
  }/*end for*/

  printf("----------tors----------\n");
  printf("ntors %d\n",ntors);
  printf("ind_glob ind_loc-> ");
  printf(" iatm1 : iatm2 : iatm3 : iatm4 |");
  printf(" latm1 : latm2 : latm3 : latm4\n");
  for (itors=1;itors<=ntors;itors++) {
      printf("%d %d-> ",tors[itors],itors);
      printf("%d : %d : %d : %d | ",tors_atom[itors][1],tors_atom[itors][2],
                                    tors_atom[itors][3],tors_atom[itors][4]);
      printf("%s : %s : %s : %s\n",maps->atm_label[tors_atom[itors][1]], 
                                   maps->atm_label[tors_atom[itors][2]],
                                   maps->atm_label[tors_atom[itors][3]],
                                   maps->atm_label[tors_atom[itors][4]]);
  }/*end for*/

  printf("----------impr----------\n");
  printf("nimpr %d\n",nimpr);
  printf("ind_glob ind_loc-> ");
  printf(" iatm1 : iatm2 : iatm3 : iatm4 |");
  printf(" latm1 : latm2 : latm3 : latm4\n");
  for (iimpr=1;iimpr<=nimpr;iimpr++) {
      printf("%d %d-> ",impr[iimpr],iimpr);
      printf("%d : %d : %d : %d | ",impr_atom[iimpr][1],impr_atom[iimpr][2],
                                    impr_atom[iimpr][3],impr_atom[iimpr][4]);
      printf("%s : %s : %s : %s\n",maps->atm_label[impr_atom[iimpr][1]], 
                                   maps->atm_label[impr_atom[iimpr][2]],
                                   maps->atm_label[impr_atom[iimpr][3]],
                                   maps->atm_label[impr_atom[iimpr][4]]);
  }/*end for*/

  printf("----------onfo----------\n");
  printf("nonfo %d\n",nonfo);
  printf("ind_glob ind_loc-> ");
  printf(" ind_atm1 : ind_atm2 |");
  printf(" lbl_atm1 : lbl_atm2\n");
  for (ionfo=1;ionfo<=nonfo;ionfo++) {
      printf("%d %d-> ",onfo[ionfo],ionfo);
      printf("%d : %d | ",onfo_atom[ionfo][1],onfo_atom[ionfo][2]);
      printf("%s : %s\n",maps->atm_label[onfo_atom[ionfo][1]], 
                         maps->atm_label[onfo_atom[ionfo][2]]);
  }/*end for*/

  printf("----------exclude----------\n");
  printf("nexclude %d\n",nexclude);
  printf("ind_glob ind_loc-> ");
  printf(" ind_atm1 : ind_atm2 |");
  printf(" lbl_atm1 : lbl_atm2\n");
  for (iexclude=1;iexclude<=nexclude;iexclude++) {
      printf("%d %d-> ",exclude[iexclude],iexclude);
      printf("%d : %d | ",exclude_atom[iexclude][1],exclude_atom[iexclude][2]);
      printf("%s : %s\n",maps->atm_label[exclude_atom[iexclude][1]], 
                         maps->atm_label[exclude_atom[iexclude][2]]);
  }/*end for*/

  printf("~~~~~~~~~~~~~~~~~~~~~~ Residue_bond ~~~~~~~~~~~~~~~~~~~~~~\n");
  getchar();
  
/*-------------------------------------------------------------------------*/
} /*end routine */
/*-------------------------------------------------------------------------*/
/*=========================================================================*/
