/***********************************************************/
/*-------------------------- C++ --------------------------*/
/***********************************************************/
/*                                                         */
/*           Some math routines for torsional transforms   */
/*                                                         */
/***********************************************************/
/*-------------------------- C++ --------------------------*/
/***********************************************************/

#include "../include/standard.h"
#include "../include/typedefs.h"
#include "../include/definitions.h"

/***********************************************************/
/*-------------------------- C++ --------------------------*/
/***********************************************************/
void d_axb_da_(double *b, double **d_axb_da)
/*---------------------------------------------------------*/
     /* calculates the derivative matrix d_axb_da  */  
     /* d_axb_da E R^(3 x 3) : output              */
     /* from                                       */
     /* b E R^3 : input                            */ 
/*---------------------------------------------------------*/
{/*begin*/
/*---------------------------------------------------------*/
d_axb_da[1][1]=0.0; d_axb_da[1][2]=b[3]; d_axb_da[1][3]=-b[2];
d_axb_da[2][1]=-b[3]; d_axb_da[2][2]=0.0; d_axb_da[2][3]=b[1];
d_axb_da[3][1]=b[2]; d_axb_da[3][2]=-b[1]; d_axb_da[3][3]=0.0;
/*---------------------------------------------------------*/
}/*end*/ 

/***********************************************************/
/*-------------------------- C++ --------------------------*/
/***********************************************************/
void d_axb_db_(double *a, double **d_axb_db)
/*---------------------------------------------------------*/
     /* calculates the derivative matrix d_axb_db   */
     /* d_axb_da E R^(3 x 3) : output               */
     /* axb is the non normalized cross product     */
     /* from                                        */
     /* a E R^3 : input                             */ 
/*---------------------------------------------------------*/
{/*begin*/
/*---------------------------------------------------------*/
d_axb_db[1][1]=0.0; d_axb_db[1][2]=-a[3]; d_axb_db[1][3]=a[2];
d_axb_db[2][1]=a[3]; d_axb_db[2][2]=0.0; d_axb_db[2][3]=-a[1];
d_axb_db[3][1]=-a[2]; d_axb_db[3][2]=a[1]; d_axb_db[3][3]=0.0;
/*---------------------------------------------------------*/
}/*end*/

/***********************************************************/
/*-------------------------- C++ --------------------------*/
/***********************************************************/
void d_axb_n_da_(double *axb, double *b, double **d_axb_n_da,
                                         double **mat1scr_3x3,
                                         double **mat2scr_3x3)
/*---------------------------------------------------------*/
     /* calculates the derivative matrix d_axb_n_da  */
     /* d_axb_n_da E R^(3 x 3) : output              */
     /* axb_n is the normalized cross product        */
     /* from                                         */
     /* b,axb E R^3 : input                          */
/*---------------------------------------------------------*/
{/*begin*/
/*---------------------------------------------------------*/
  /* construct v = a x b  */
  int i,j;
  int ndim = 3;
  double** axb_axb_n = mat1scr_3x3; 
  double** d_axb_da = mat2scr_3x3;
  /* norm of axb  */
  double norm = sqrt(axb[1]*axb[1]+axb[2]*axb[2]+axb[3]*axb[3]); 
  /* construct axb_axb_n  */
  for (i=1;i<=ndim;i++) {
      for (j=1;j<=ndim;j++) {
          axb_axb_n[i][j] = -axb[i]*axb[j]/(norm*norm*norm);
          if ( i == j ) 
             axb_axb_n[i][i] += 1.0/norm;  
      }
  }   
  /* construct d_axb_da  */
  d_axb_da_(b,d_axb_da);
  /* construct d_axb_da_n  */
  mat_mat(axb_axb_n,d_axb_da,d_axb_n_da,ndim);
  /* free matrices  */
/*---------------------------------------------------------*/
}/*end*/   

/***********************************************************/
/*-------------------------- C++ --------------------------*/
/***********************************************************/
void d_axb_n_db_(double *axb, double *a, double **d_axb_n_db,
                                         double **mat1scr_3x3,
                                         double **mat2scr_3x3)
/*---------------------------------------------------------*/
     /* calculates the derivative matrix d_axb_n_db  */  
     /* d_axb_n_db E R^(3 x 3) : output              */
     /* axb_n is the normalized cross product        */
     /* from                                         */
     /* a,axb E R^3 : input                          */ 
/*---------------------------------------------------------*/
{/*begin*/
/*---------------------------------------------------------*/
  /* construct v = a x b  */
  int i,j;
  int ndim = 3;
  double** axb_axb_n = mat1scr_3x3;
  double** d_axb_db = mat2scr_3x3;
  /* norm of axb */
  double norm = sqrt(axb[1]*axb[1]+axb[2]*axb[2]+axb[3]*axb[3]); 
  /* construct axb_axb_n */
  for (i=1;i<=ndim;i++) {
      for (j=1;j<=ndim;j++) {
          axb_axb_n[i][j] = -axb[i]*axb[j]/(norm*norm*norm);
          if ( i == j ) 
             axb_axb_n[i][i] += 1.0/norm;  
      }
  }   
  /* construct d_axb_db  */
  d_axb_db_(a,d_axb_db);
  /* construct d_axb_db_n */
  mat_mat(axb_axb_n,d_axb_db,d_axb_n_db,ndim);
  /* free matrices */
/*---------------------------------------------------------*/
}/*end*/   
  
/*---------------------------------------------------------*/
/***********************************************************/
/*-------------------------- C++ --------------------------*/
/***********************************************************/
void dR_dr_(double **dex_dr, double **dey_dr, double **dez_dr, 
           double **dR_drx, double **dR_dry, double **dR_drz)
/*---------------------------------------------------------*/
     /* Construct the tensor dR/dr                   */
     /* where R E R^(3 x 3), r E R^3                 */
     /* in the form                                  */
     /* dR/drx, dR/dry, dR/drz E R^(3 x 3): output   */ 
     /* from                                         */
     /* dex/dr, dey/dr, dez/dr E R^(3 x 3): input    */ 
     /* where                                        */
     /*     |  ex  |                                 */ 
     /* R = |  ey  |                                 */
     /*     |  ez  |                                 */
     /* where ex, ey, ez E R^(1 x 3)                 */
/*---------------------------------------------------------*/
{/*begin*/
 int ndim = 3;
 int j;
     for (j=1;j<=ndim;j++) {
         dR_drx[1][j] = dex_dr[j][1];
         dR_drx[2][j] = dey_dr[j][1];
         dR_drx[3][j] = dez_dr[j][1];
     }
     for (j=1;j<=ndim;j++) {
         dR_dry[1][j] = dex_dr[j][2];
         dR_dry[2][j] = dey_dr[j][2];
         dR_dry[3][j] = dez_dr[j][2];
     }
     for (j=1;j<=ndim;j++) {
         dR_drz[1][j] = dex_dr[j][3];
         dR_drz[2][j] = dey_dr[j][3];
         dR_drz[3][j] = dez_dr[j][3];
     }
/*---------------------------------------------------------*/
}/*end*/      

/***********************************************************/
/*-------------------------- C++ --------------------------*/
/***********************************************************/
void dRT_dr_(double **dex_dr, double **dey_dr, double **dez_dr,
          double **dRT_drx, double **dRT_dry, double **dRT_drz)
/*---------------------------------------------------------*/
     /* Construct the tensor dRT/dr                   */
     /* where RT E R^(3 x 3), r E R^3                 */
     /* in the form                                   */
     /* dRT/drx, dRT/dry, dRT/drz E R^(3 x 3): output */ 
     /* from                                          */
     /* dex/dr, dey/dr, dez/dr E R^(3 x 3): input     */ 
     /* where                                         */
     /*     |  ex  |                                  */
     /* R = |  ey  |                                  */
     /*     |  ez  |                                  */
     /* where ex, ey, ez E R^(1 x 3)                  */
/*---------------------------------------------------------*/
{/*begin*/
 int ndim = 3;
 int j;
     for (j=1;j<=ndim;j++) {
         dRT_drx[j][1] = dex_dr[j][1];
         dRT_drx[j][2] = dey_dr[j][1];
         dRT_drx[j][3] = dez_dr[j][1];
     }
     for (j=1;j<=ndim;j++) {
         dRT_dry[j][1] = dex_dr[j][2];
         dRT_dry[j][2] = dey_dr[j][2];
         dRT_dry[j][3] = dez_dr[j][2];
     }
     for (j=1;j<=ndim;j++) {
         dRT_drz[j][1] = dex_dr[j][3];
         dRT_drz[j][2] = dey_dr[j][3];
         dRT_drz[j][3] = dez_dr[j][3];
     }
/*---------------------------------------------------------*/
}/*end*/      

/***********************************************************/

/***********************************************************/
/*-------------------------- C++ --------------------------*/
/***********************************************************/
void de_dra_(double *ra, double *rb, double *e, double **de_dra) 
/*---------------------------------------------------------*/
     /* Construct the matrix de_dra E R^(3 x 3) : output */ 
     /* where ra,rb,e E R^3  : input                     */
     /* and e = (ra-rb)/||ra-rb||                        */
     /* norm : ||ra-rb||                                 */
/*--------------------------------------------------------*/
{/*begin*/
 int i,j;
 int ndim = 3;
 double norm;

 if (e[1] != 0.0) {norm = (ra[1]-rb[1])/e[1];} 
 else { 
      if (e[2] != 0.0) {norm = (ra[2]-rb[2])/e[2];} 
      else {
           if (e[3] != 0.0) {norm = (ra[3]-rb[3])/e[3];}
           else {nrerror("de_dra_ : ra = rb");}
      } 
 } 

 for (i=1;i<=ndim;i++) {
     for (j=1;j<=ndim;j++) {
         de_dra[i][j] = -e[j]*e[i];
         if ( i == j ) 
            de_dra[i][j] += 1.0;
         de_dra[i][j] /= norm;
     }
 } 
/*---------------------------------------------------------*/
}/*end*/ 

/***********************************************************/
/*-------------------------- C++ --------------------------*/
/***********************************************************/
void de_drb_(double *ra, double *rb, double *e, double **de_drb) 
/*---------------------------------------------------------*/
 /* Construct the matrix de_drb E R^(3 x 3) : output 
 /* where ra,rb,e E R^3  : input 
 /* and e = (ra-rb)/||ra-rb||
 /* norm : ||ra-rb|| 
/*---------------------------------------------------------*/
{/*begin*/
 int i,j;
 int ndim = 3;
 double norm;

 if (e[1] != 0.0) {norm = (ra[1]-rb[1])/e[1];} 
 else { 
      if (e[2] != 0.0) {norm = (ra[2]-rb[2])/e[2];} 
      else {
           if (e[3] != 0.0) {norm = (ra[3]-rb[3])/e[3];}
           else {nrerror("de_drb_ : ra = rb");}
      } 
 } 


 for (i=1;i<=ndim;i++) {
     for (j=1;j<=ndim;j++) {
         de_drb[i][j] = e[j]*e[i];
         if ( i == j ) 
            de_drb[i][j] -= 1.0;
         de_drb[i][j] /= norm;
     }
 } 
/*---------------------------------------------------------*/
}/*end*/ 

/***********************************************************/
/*-------------------------- C++ --------------------------*/
/***********************************************************/
void dRa_dr_(double **dR_drx, double **dR_dry, double **dR_drz,
             double *a, double** dRa_dr, double *vecscr_3x1) 
/*---------------------------------------------------------*/
 /* Construct the matrix dRa_dr E R^(3 x 3) : output 
 /* which is d(R(r)*a)/dr
 /* where R E R^(3 x 3), a,r E R^3 
 /* dR_dri = dR/dri E R^(3 x 3), i={x,y,z} : input
 /* a E R^3 : input  
/*---------------------------------------------------------*/
{/*begin*/
 int i;
 int ndim = 3;
 double *b = vecscr_3x1;
/*---------------------------------------------------------*/
 /* b = dR_drx*a  */
 mat_vec(dR_drx,a,b,ndim);
/*---------------------------------------------------------*/
 /* put b as the first column of dRa_dr  */
 for (i=1;i<=ndim;i++)  
     dRa_dr[i][1] = b[i];  
/*---------------------------------------------------------*/
 /* b = dR_dry*a  */
 mat_vec(dR_dry,a,b,ndim);
/*---------------------------------------------------------*/
 /* put b as the second column of dRa_dr  */
 for (i=1;i<=ndim;i++)  
     dRa_dr[i][2] = b[i];  
/*---------------------------------------------------------*/
 /* b = dR_drz*a  */
 mat_vec(dR_drz,a,b,ndim);
/*---------------------------------------------------------*/
 /* put b as the third column of dRa_dr  */
 for (i=1;i<=ndim;i++)  
     dRa_dr[i][3] = b[i]; 
/*---------------------------------------------------------*/
 /* free dvector  */
}/*end*/ 
