#include "../include/standard.h"
#include "../include/typedefs.h"
#include "../include/definitions.h"
#include "../include/utillib.h"
#include "../include/cmalloc.h"

#define DEBUG
#undef DEBUG

/*=========================================================================*/
void Pos::move_entry(System &system, int index, double sig_trans)
/*-------------------------------------------------------------------------*/
/*  This routine moves a single entry                                      */
/*  Degrees of freedom: three translations                                 */
{ /*begin routine */
/*-------------------------------------------------------------------------*/
  
  double T0    =  system.general.T;            /* Repli.0 temperature in K */
  double T     =  (this->T)*BOLTZ;             /* Replica temperature in K */
  double sTpT0 =  sqrt(T/T0);

  long *seed  = &(system.general.seed);
  double one  = 1.0;

  double sig, dx, dy, dz;

/*-------------------------------------------------------------------------*/
/* I) Propagate in Normal distribution with zero mean                      */
/*      f(x) = 1/(sig*sqrt(2Pi)) exp(-x^2/(2 sig^2))                       */
/*      double mb(1,sig,seed) from f(x)                                    */

  sig  = sTpT0*sig_trans;

  dx  =  mb(one,sig,seed);
  dy  =  mb(one,sig,seed);
  dz  =  mb(one,sig,seed);

  xi[index] += dx; 
  yi[index] += dy;
  zi[index] += dz;

/*-------------------------------------------------------------------------*/
} /*end routine */
/*=========================================================================*/


/*=========================================================================*/
void Pos::move_entry(System &system,int i1,int i2,int i3,double sig_trans,
                                                         double sig_rot)
/*-------------------------------------------------------------------------*/
/*  This routine moves a three point entry                                 */
/*  Degrees of freedom: three translations (rigid)                         */
/*                      three rotations    (rigid)                         */
/*                      bend               (internal)                      */
{ /*begin routine */
/*-------------------------------------------------------------------------*/

  int i, ndim;
  double norm, norm_ex;
  double sig, dx, dy, dz;
  double e1_ry_exp, zero;

  double phi, psi, theta;
  double sig_rot_scale;

  double T0    =  system.general.T;            /* Repli.0 temperature in K */
  double T     =  (this->T)*BOLTZ;             /* Replica temperature in K */
  double sTpT0 =  sqrt(T/T0);

  long *seed  = &(system.general.seed);
  double one  = 1.0;

  double**  A = system.tors_defs.rotate_utils.A;
  double**  B = system.tors_defs.rotate_utils.B;
  double**  C = system.tors_defs.rotate_utils.C;
  double**  R = system.tors_defs.rotate_utils.R;
  double** RT = system.tors_defs.rotate_utils.RT;

  double*  r1 = system.tors_defs.rotate_utils.r1;
  double*  r2 = system.tors_defs.rotate_utils.r2;
  double*  r3 = system.tors_defs.rotate_utils.r3;
  double*  r4 = system.tors_defs.rotate_utils.r4;

  double*  ex = system.tors_defs.rotate_utils.ex;
  double*  ey = system.tors_defs.rotate_utils.ey;
  double*  ez = system.tors_defs.rotate_utils.ez;
  double*  e1 = system.tors_defs.rotate_utils.e1;
  double*  e2 = system.tors_defs.rotate_utils.e2;

  double*  e1_r = system.tors_defs.rotate_utils.a;
  double*  e2_r = system.tors_defs.rotate_utils.b;

  double*  e1_Ar = ex;                      /* use after ex is stored in R */
  double*  e2_Ar = ey;                      /* use after ey is stored in R */

/*-------------------------------------------------------------------------*/
/* I) Propagate in Normal distribution with zero mean                      */
/*      f(x) = 1/(sig*sqrt(2Pi)) exp(-x^2/(2 sig^2))                       */
/*      double mb(1,sig,seed) from f(x)                                    */

/*-------------------------------------------------------------------------*/
/* I.A) Translational propagation                                          */

  sig  = sTpT0*sig_trans;

  dx  =  mb(one,sig,seed);
  dy  =  mb(one,sig,seed);
  dz  =  mb(one,sig,seed);

  xi[i1] += dx; xi[i2] += dx; xi[i3] += dx;
  yi[i1] += dy; yi[i2] += dy; yi[i3] += dy;
  zi[i1] += dz; zi[i2] += dz; zi[i3] += dz;

/*-------------------------------------------------------------------------*/
/* I.B) Rotational propagation                                             */
/* a) prior to rotation an internal reference frame is constructed         */
/* b) move into an internal frame                                          */ 
/* c) apply rigid body rotation                                            */
/* d) move back from internal frame                                        */
/*         rotation matrices : R into the internal frame                   */ 
/*                             RT from the internal frame                  */ 
/* DEBUG on an artifial symmetric case                                     */

  r1[1] = xi[i1]; r1[2] = yi[i1]; r1[3] = zi[i1];
  r2[1] = xi[i2]; r2[2] = yi[i2]; r2[3] = zi[i2];
  r3[1] = xi[i3]; r3[2] = yi[i3]; r3[3] = zi[i3];

#ifdef DEBUG
  r1[1] = 0.0; r1[2] =  0.5; r1[3] = 0.0;
  r2[1] = 0.0; r2[2] =  0.0; r2[3] = 0.5;
  r3[1] = 0.0; r3[2] = -0.5; r3[3] = 0.0;
#endif

  r4[1] = r1[1] + (r3[1] - r1[1])/2.0;
  r4[2] = r1[2] + (r3[2] - r1[2])/2.0;
  r4[3] = r1[3] + (r3[3] - r1[3])/2.0;


/*-------------------------------------------------------------------------*/
/* a) create reference frame                                               */
					 
  ex[1] = r4[1] - r2[1];   e1[1] = r1[1] - r2[1];   e2[1] = r3[1] - r2[1];  
  ex[2] = r4[2] - r2[2];   e1[2] = r1[2] - r2[2];   e2[2] = r3[2] - r2[2];
  ex[3] = r4[3] - r2[3];   e1[3] = r1[3] - r2[3];   e2[3] = r3[3] - r2[3];

  norm = sqrt(ex[1]*ex[1]+ex[2]*ex[2]+ex[3]*ex[3]);
  ex[1] /= norm; ex[2] /= norm; ex[3] /= norm;
  norm_ex = norm;

  cprd(e1[1],e1[2],e1[3],e2[1],e2[2],e2[3],&(ez[1]),&(ez[2]),&(ez[3]));
  norm = sqrt(ez[1]*ez[1]+ez[2]*ez[2]+ez[3]*ez[3]);
  ez[1] /= norm; ez[2] /= norm; ez[3] /= norm;

  cprd(ex[1],ex[2],ex[3],ez[1],ez[2],ez[3],&(ey[1]),&(ey[2]),&(ey[3]));


/*-------------------------------------------------------------------------*/
/* b) move into the frame                                                  */

  ndim = 3;
  RT[1][1]=ex[1]; RT[1][2]=ey[1]; RT[1][3]=ez[1];
  RT[2][1]=ex[2]; RT[2][2]=ey[2]; RT[2][3]=ez[2];
  RT[3][1]=ex[3]; RT[3][2]=ey[3]; RT[3][3]=ez[3];
  mat_trn(RT,R,ndim);   
 
  mat_vec(R,e1,e1_r,ndim);       
  mat_vec(R,e2,e2_r,ndim);    


/*-------------------------------------------------------------------------*/
/* c) apply rigid body rotation                                            */
  
  sig = sTpT0*sig_rot;

  //pitch-roll-yaw convention
   phi   =  mb(one,sig,seed);
   psi   =  mb(one,sig,seed);
   theta =  mb(one,sig,seed);

   A[1][1] = cos(theta)*cos(phi);
   A[1][2] = cos(theta)*sin(phi);
   A[1][3] = -sin(theta);
   A[2][1] = sin(psi)*sin(theta)*cos(phi)-cos(psi)*sin(phi);
   A[2][2] = sin(psi)*sin(theta)*sin(phi)+cos(psi)*cos(phi);
   A[2][3] = cos(theta)*sin(psi);
   A[3][1] = cos(psi)*sin(theta)*cos(phi)+sin(psi)*sin(phi);
   A[3][2] = cos(psi)*sin(theta)*sin(phi)-sin(psi)*cos(phi);
   A[3][3] = cos(theta)*cos(psi);

  mat_vec(A,e1_r,e1_Ar,ndim);
  mat_vec(A,e2_r,e2_Ar,ndim);

/*-------------------------------------------------------------------------*/
/* d) move back from internal frame                                        */

  mat_vec(RT,e1_Ar,e1,ndim);    
  mat_vec(RT,e2_Ar,e2,ndim);    

  r1[1] = r2[1] + e1[1];  r3[1] = r2[1] + e2[1];
  r1[2] = r2[2] + e1[2];  r3[2] = r2[2] + e2[2];
  r1[3] = r2[3] + e1[3];  r3[3] = r2[3] + e2[3];

  xi[i1] = r1[1]; yi[i1] = r1[2]; zi[i1] = r1[3];
  xi[i3] = r3[1]; yi[i3] = r3[2]; zi[i3] = r3[3];

/*-------------------------------------------------------------------------*/
/* II) DEBUG                                                               */
  
#ifdef DEBUG
   zero = 0.0;
   e1_ry_exp = sqrt( (r1[1]-r3[1])*(r1[1]-r3[1])+ 
                     (r1[2]-r3[2])*(r1[2]-r3[2])+
                     (r1[3]-r3[3])*(r1[3]-r3[3]) )/2.0;
   printf("e1_rx_exp =  e2_rx_exp = %g\n",norm_ex);
   printf("e1_ry_exp = -e2_ry_exp = %g\n",e1_ry_exp);
   printf("e1_rz_exp =  e2_rz_exp = %g\n",zero);
   printf("e1_rx %g, e1_ry %g, e1_rz %g\n",e1_r[1],e1_r[2],e1_r[3]);
   printf("e2_rx %g, e2_ry %g, e2_rz %g\n",e2_r[1],e2_r[2],e2_r[3]);
   getchar();
#endif

/*-------------------------------------------------------------------------*/
} /*end routine */
/*=========================================================================*/

