#include "../include/standard.h"
#include "../include/typedefs.h"
#include "../include/definitions.h"
#include "../include/utillib.h"
#include "../include/cmalloc.h"

#define DEBUG
#undef DEBUG

/*=========================================================================*/
void Pos::rotate_entry(System &system, double *ex, double *ey, 
                                       double *ez, double *e0, 
                                       double psi_x, double theta_y,
                                       double phi_z, int i1, int i2, int i3)
/*-------------------------------------------------------------------------*/
/*  This routine rotates a single entry defined by three atoms             */
/*  with respect to the reference frame with origin e0 and orthogonal      */
/*  unit vecotrs ex, ey, ez. The magnitude or rotation is defined by       */
/*  psi_x (around ex), theta_y (around ey) and phi_z (around ez).          */
/*                                                                         */
/*  ex  Lth: 3, 3D unit vecor ex[1] = ex_x, ex[2] = ex_y, ex[3] = ex_z     */
/*  ey  Lth: 3, 3D unit vecor ey[1] = ey_x, ey[2] = ey_y, ey[3] = ey_z     */
/*  ez  Lth: 3, 3D unit vecor ez[1] = ez_x, ez[2] = ez_y, ez[3] = ez_z     */
/*  phi_x, theta_y, phi_z : desired rotational displacements               */
/*  i1, i2, i3: indices of the three atoms                                 */
{ /*begin routine */
/*-------------------------------------------------------------------------*/

  int i, ndim;
  double psi,theta, phi;

  double**  A = system.tors_defs.rotate_utils.A;
  double**  B = system.tors_defs.rotate_utils.B;
  double**  C = system.tors_defs.rotate_utils.C;
  double**  R = system.tors_defs.rotate_utils.R;
  double** RT = system.tors_defs.rotate_utils.RT;
  double** I  = system.tors_defs.rotate_utils.I;

  double*  r1 = system.tors_defs.rotate_utils.r1;
  double*  r2 = system.tors_defs.rotate_utils.r2;
  double*  r3 = system.tors_defs.rotate_utils.r3;

  double*  e1 = system.tors_defs.rotate_utils.e1;
  double*  e2 = system.tors_defs.rotate_utils.e2;
  double*  e3 = system.tors_defs.rotate_utils.e3;

  double*  e1_r = system.tors_defs.rotate_utils.e1_r;
  double*  e2_r = system.tors_defs.rotate_utils.e2_r;
  double*  e3_r = system.tors_defs.rotate_utils.e3_r;

  double*  e1_Ar = system.tors_defs.rotate_utils.e1_Ar;
  double*  e2_Ar = system.tors_defs.rotate_utils.e2_Ar;
  double*  e3_Ar = system.tors_defs.rotate_utils.e3_Ar;

/*-------------------------------------------------------------------------*/
/* I.B) Rotational propagation                                             */
/* a) translate into internal frame                                        */
/* b) rotate into internal frame                                           */ 
/* c) apply rigid body rotation                                            */
/* d) rotate from internal frame                                           */
/* e) translate from internal frame                                        */
/*         rotation matrices : R into the internal frame                   */ 
/*                             RT from the internal frame                  */ 
/* DEBUG on an artifial symmetric case                                     */

  r1[1] = xi[i1]; r1[2] = yi[i1]; r1[3] = zi[i1];
  r2[1] = xi[i2]; r2[2] = yi[i2]; r2[3] = zi[i2];
  r3[1] = xi[i3]; r3[2] = yi[i3]; r3[3] = zi[i3];

  
#ifdef DEBUG
   printf(" before rotate \n");
   printf("i1 %d, i2 %d, i3 %d\n",i1,i2,i3);
   printf("r1[1] %g, r1[2] %g, r1[3] %g\n",r1[1],r1[2],r1[3]);
   printf("r2[1] %g, r2[2] %g, r2[3] %g\n",r2[1],r2[2],r2[3]);
   printf("r3[1] %g, r3[2] %g, r3[3] %g\n",r3[1],r3[2],r3[3]);
   getchar();
#endif


/*-------------------------------------------------------------------------*/
/* a) translate into internal frame                                        */
					   
  e1[1] = r1[1] - e0[1];   e2[1] = r2[1] - e0[1];   e3[1] = r3[1] - e0[1];  
  e1[2] = r1[2] - e0[2];   e2[2] = r2[2] - e0[2];   e3[2] = r3[2] - e0[2];  
  e1[3] = r1[3] - e0[3];   e2[3] = r2[3] - e0[3];   e3[3] = r3[3] - e0[3];  

/*-------------------------------------------------------------------------*/
/* b) move into the frame                                                  */

  ndim = 3;
  RT[1][1]=ex[1]; RT[1][2]=ey[1]; RT[1][3]=ez[1];
  RT[2][1]=ex[2]; RT[2][2]=ey[2]; RT[2][3]=ez[2];
  RT[3][1]=ex[3]; RT[3][2]=ey[3]; RT[3][3]=ez[3];
  mat_trn(RT,R,ndim);   
 
  mat_vec(R,e1,e1_r,ndim);       
  mat_vec(R,e2,e2_r,ndim);    
  mat_vec(R,e3,e3_r,ndim);    

#ifdef DEBUG 
  //  mat_vec(I,e1,e1_r,ndim);       
  //  mat_vec(I,e2,e2_r,ndim);    
  //  mat_vec(I,e3,e3_r,ndim);
  //  mat_mat(R,RT,A,ndim);
  //  mat_out(A,ndim);
  //  getchar();
#endif

/*-------------------------------------------------------------------------*/
/* c) apply rigid body rotation                                            */
/*    pitch-roll-yaw convention                                            */

   phi   =  phi_z;
   psi   =  psi_x;
   theta =  theta_y;

   A[1][1] = cos(theta)*cos(phi);
   A[1][2] = cos(theta)*sin(phi);
   A[1][3] = -sin(theta);
   A[2][1] = sin(psi)*sin(theta)*cos(phi)-cos(psi)*sin(phi);
   A[2][2] = sin(psi)*sin(theta)*sin(phi)+cos(psi)*cos(phi);
   A[2][3] = cos(theta)*sin(psi);
   A[3][1] = cos(psi)*sin(theta)*cos(phi)+sin(psi)*sin(phi);
   A[3][2] = cos(psi)*sin(theta)*sin(phi)-sin(psi)*cos(phi);
   A[3][3] = cos(theta)*cos(psi);

  mat_vec(A,e1_r,e1_Ar,ndim);
  mat_vec(A,e2_r,e2_Ar,ndim);
  mat_vec(A,e3_r,e3_Ar,ndim);

#ifdef DEBUG
  //  mat_vec(I,e1_r,e1_Ar,ndim);
  //  mat_vec(I,e2_r,e2_Ar,ndim);
  //  mat_vec(I,e3_r,e3_Ar,ndim);
#endif

/*-------------------------------------------------------------------------*/
/* d) rotate from internal frame                                           */

  mat_vec(RT,e1_Ar,e1,ndim);    
  mat_vec(RT,e2_Ar,e2,ndim);    
  mat_vec(RT,e3_Ar,e3,ndim);    

#ifdef DEBUG 
  //  mat_vec(I,e1_Ar,e1,ndim);       
  //  mat_vec(I,e2_Ar,e2,ndim);    
  //  mat_vec(I,e3_Ar,e3,ndim);    
#endif


/*-------------------------------------------------------------------------*/
/* e) translate from internal frame                                        */

  r1[1] = e1[1] + e0[1];  r2[1] =  e2[1] + e0[1];   r3[1] = e3[1] + e0[1];  
  r1[2] = e1[2] + e0[2];  r2[2] =  e2[2] + e0[2];   r3[2] = e3[2] + e0[2];  
  r1[3] = e1[3] + e0[3];  r2[3] =  e2[3] + e0[3];   r3[3] = e3[3] + e0[3];  

  xi[i1] = r1[1]; yi[i1] = r1[2]; zi[i1] = r1[3];
  xi[i2] = r2[1]; yi[i2] = r2[2]; zi[i2] = r2[3];
  xi[i3] = r3[1]; yi[i3] = r3[2]; zi[i3] = r3[3];

/*-------------------------------------------------------------------------*/
/* II) DEBUG                                                               */
  
#ifdef DEBUG
   printf(" after rotate \n");
   printf("r1[1] %g, r1[2] %g, r1[3] %g\n",r1[1],r1[2],r1[3]);
   printf("r2[1] %g, r2[2] %g, r2[3] %g\n",r2[1],r2[2],r2[3]);
   printf("r3[1] %g, r3[2] %g, r3[3] %g\n",r3[1],r3[2],r3[3]);
   getchar();
#endif

/*-------------------------------------------------------------------------*/
} /*end routine */
/*=========================================================================*/

