#include "../../include/standard.h" 
#include "../../include/typedefs.h"
#include "../../include/definitions.h"
#include "../../include/utillib.h"
#include "../../include/cmalloc.h"

#define DEBUG
#undef DEBUG

/*=========================================================================*/
 void Tors_defs::setup_cgresmodel_1pt(const Maps *maps, const Pos *pos, 
                                                       const General *general)
/*-------------------------------------------------------------------------*/
/*  Sets up all the torsional angles for cgres_model_3pt                   */
/*-------------------------------------------------------------------------*/
{/*begin routine*/  
/*-------------------------------------------------------------------------*/

  int ires, itrans;
  int iatm, iiatm, ibranch;  
  char   err_msg[LINE_NCHARMAX];


  int    nmol  =  maps->nmol;

/*-------------------------------------------------------------------------*/
/* 0) Only one molecule is supported                                       */

  if (nmol!=1) {
    sprintf(err_msg,"Tors_defs::setup_cgresmodel_1pt:nmol!=1:not supported");
    errorMessage(err_msg);
  }/*endif*/

/*-------------------------------------------------------------------------*/
/* I) Get the number of transformations                                    */
/*    Assume 1 transformations/residue, except for first three residue     */

  num_trans = (pos->nres)-3;

/*-------------------------------------------------------------------------*/
/* II) Allocate variables                                                  */

  res_id         =  new int[num_trans]-1;
  ityp_trans     =  new int[num_trans]-1;
  off_trans      =  new int[num_trans]-1;
  natm_trans     =  new int[num_trans]-1;

  ind_trans      =  cmall_int_mat(1, num_trans, 1, NATM_TRANS_MAX);
  ind_prim_tors  =  cmall_int_mat(1, num_trans, 1, 4);

/*-------------------------------------------------------------------------*/
/* III) Define the group of atoms to be transformed                        */
/*      1) Loop over the residues having groups (first excluded)           */
/*      2) The only atom in residue is CA branch_root_f[0][0]              */
/*      3) Assume one groups/residue                                       */
/*      4) Group 1 : branch_root_f[0][0]                                   */

  itrans = 0;
  for (ires=4;ires<=pos->nres;ires++) {
       itrans++;
       if (itrans>num_trans) {
        sprintf(err_msg,"Tors_defs::setup_cgresmodel_CA:itrans>num_trans");
        errorMessage(err_msg);
       }/*endif*/
       res_id[itrans]     = ires;
       natm_trans[itrans] = 1;
       ind_trans[itrans][1]     =  pos->residue[ires].branch_root_f[0][0];
       ind_prim_tors[itrans][4] =  pos->residue[ires].branch_root_f[0][0];
       ind_prim_tors[itrans][3] =  pos->residue[ires-1].branch_root_f[0][0];
       ind_prim_tors[itrans][2] =  pos->residue[ires-2].branch_root_f[0][0];
       ind_prim_tors[itrans][1] =  pos->residue[ires-3].branch_root_f[0][0];
  }/*endfor*/

  if (itrans!=num_trans) {
   sprintf(err_msg,"In Tors_defs::setup_cgresmodel_CA:itrans!=num_trans");
   errorMessage(err_msg);
  }/*endif*/

/*-------------------------------------------------------------------------*/
/* IV) Transformation types                                                */
/*     By default, everybody is transformed                                */

  for (itrans=1;itrans<=num_trans;itrans++) {
       ityp_trans[itrans] = 1;
       off_trans[itrans]  = 0;
  }/*end for*/

/*-------------------------------------------------------------------------*/
/* V) If stride is defined in .pdb, only sample                            */
/*       along 'C' secondary structure type residue                        */   

  if (general->stride_defined==1) {
    if (nstride!=1) {
      sprintf(err_msg, "Tors_defs::setup_cgresmodel_1pt: nstride!=1");
      errorMessage(err_msg);
    }/*endif*/
    if (strlen(&(stride[1][0]))-1!=pos->nres) {
      sprintf(err_msg, "Tors_defs::setup_cgresmodel_1pt: corrupted STRIDE");
      errorMessage(err_msg);
    }/*endif*/
    for (itrans=1;itrans<=num_trans;itrans++) {
         ityp_trans[itrans] = 1;
         off_trans[itrans]  = 1;
         if (stride[1][res_id[itrans]] == 'C') {
             off_trans[itrans] = 0;
         }/*endif*/
    }/*end for*/
  }/*end if*/


/*-------------------------------------------------------------------------*/
/* VI) Adjust transformation types                                         */
/*     if off_trans == 1 => ityp_trans = 0 (identical)                     */
/*     if off_trans == 0 => ityp_trans = 1 (prop Gauss)                    */

  for (itrans=1;itrans<=num_trans;itrans++) {
    if (off_trans[itrans] == 1) {ityp_trans[itrans] = 0;}
    if (off_trans[itrans] == 0) {ityp_trans[itrans] = 1;}
  }/*end for*/


/*-------------------------------------------------------------------------*/
/* VII) Setup rotate_utils                                                  */

  rotate_utils.setup();
  
/*-------------------------------------------------------------------------*/
/* VIII) Debug                                                               */

#ifdef DEBUG
  printf("############################ All Torsions ############################\n");
  write_alldata(maps);
#endif

/*-------------------------------------------------------------------------*/
}/*end routine*/  
/*=========================================================================*/

