#include "../../include/standard.h"
#include "../../include/typedefs.h"
#include "../../include/definitions.h"
#include "../../include/utillib.h"
#include "../../include/cmalloc.h"

#define DEBUG
#undef DEBUG

/*=========================================================================*/
 void Tors_defs::setup_3pt_cblc(const Maps *maps, const Pos *pos,
                                                  const General *general)
/*-------------------------------------------------------------------------*/
/*  Sets up all the torsional angles for cgres_model_3pt                   */
/*-------------------------------------------------------------------------*/
{/*begin routine*/
/*-------------------------------------------------------------------------*/

  int ires, itrans, iclbase, iclos;
  int iatm, iiatm, ibranch;

  double dx, dy, dz;
  double *x = pos->x;
  double *y = pos->y;
  double *z = pos->z;

  int odd, even;
  int ind_CA, ind_O, ind_CMA;
  int ind_prevres_CA, ind_prevres_O, ind_prevres_CMA;
  int ind_nextres_CA, ind_nextres_O, ind_nextres_CMA;

  char  *atm_label;
  char   err_msg[LINE_NCHARMAX];

  int    nmol  =  maps->nmol;

/*-------------------------------------------------------------------------*/
/* 0) Only one molecule is supported                                       */

  if (nmol!=1) {
    sprintf(err_msg,"Tors_defs::setup_cgresmodel_3pt_cblc\n"
                    ":nmol!=1:not supported");
    errorMessage(err_msg);
  }/*endif*/

/*-------------------------------------------------------------------------*/
/* I) Get the number of transformations, closures and closure bases        */
/*    It depends if there are odd or even number of residues               */

  odd = 0; even = 0;
  if ( ((pos->nres) % 2) == 0 ) {even = 1;}
  else {odd = 1;}

  if (even == 1) {
      num_trans  = (pos->nres) - 1;
      num_clbase = (pos->nres)/2;
      num_clos   =  num_clbase - 1;
  }/*endif*/

  if (odd == 1) {
      num_trans  = (pos->nres) - 1;
      num_clbase = ((pos->nres)- 1)/2;
      num_clos   =  num_clbase - 1;         
  }/*endif*/


/*-------------------------------------------------------------------------*/
/* II)                                                                     */
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* II.A) Allocate variables                                                */

  res_id         =  new int[num_trans]-1;
  ityp_trans     =  new int[num_trans]-1;
  off_trans      =  new int[num_trans]-1;
  natm_trans     =  new int[num_trans]-1;
  ind_clos_trans =  new int[num_trans]-1;
  prop_sig_scale =  new double[num_trans]-1;

  natm_clos      =  new int[num_clos]-1;
  ityp_clos      =  new int[num_clos]-1;

  natm_clbase     =  new int[num_clbase]-1;
  ind_clbase_free =  new int[num_clbase]-1;
  res_id_clbase   =  new int[num_clbase]-1;
  freeze_clbase   =  new int[num_clbase]-1;

  ind_trans      =  cmall_int_mat(1, num_trans, 1, NATM_TRANS_MAX);
  ind_prim_tors  =  cmall_int_mat(1, num_trans, 1, 4);
  ind_atm_clos   =  cmall_int_mat(1, num_clos, 1, NATM_CLOS_MAX);
  ind_atm_clbase =  cmall_int_mat(1, num_clbase, 1, NATM_CLBASE_MAX);
  const_clos     =  cmall_mat(1, num_clos, 1, NATM_CLOS_MAX);

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* II.B) Initialize default variables                                      */

  for (itrans=1;itrans<=num_trans;itrans++) {
       ityp_trans[itrans]     = 1;
       off_trans[itrans]      = 0;
       ind_clos_trans[itrans] = 0;
       prop_sig_scale[itrans] = 1.0;
  }/*endfor*/

  num_clbase_free = num_clbase;
  for (iclbase=1;iclbase<=num_clbase;iclbase++) {
       freeze_clbase[iclbase] = 0;
       ind_clbase_free[iclbase] = iclbase;
  }/*endfor*/

/*-------------------------------------------------------------------------*/
/* III.A)  Decide if 3pt protein or nucleic acid                           */
  int nucleic = 0;
  int protein = 0;
  iatm = pos->residue[1].branch_root_f[0][0];
  if (strcasecmp(maps->upac_atm_label[iatm],"CA")==0){protein = 1;}  
  if (strcasecmp(maps->upac_atm_label[iatm],"P")==0) {nucleic = 1;}
#ifdef DEBUG
      if (protein) {printf("This is a protein in 3pt representation\n");}
      if (nucleic) {printf("This is a nucleic acid in 3pt representation\n");}
      getchar();
#endif

/*-------------------------------------------------------------------------*/
/* III.B) Define the group of atoms to be transformed for proteins         */

 if (protein) {

  itrans  = 0;
  iclbase = 0;
  iclos   = 0;
  for (ires=1;ires<=pos->nres;ires++) {
/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/
/* III.C.1) Find atom indices                                              */

       for (iatm=1;iatm<=pos->residue[ires].natom;iatm++) {
            iiatm     =  pos->residue[ires].atom[iatm];
            atm_label =  maps->upac_atm_label[iiatm];

            if (strcasecmp(atm_label,"CA")==0)  {ind_CA  = iiatm;}
            if (strcasecmp(atm_label,"O")==0)   {ind_O   = iiatm;}
            if (strcasecmp(atm_label,"CMA")==0) {ind_CMA = iiatm;}
       }/*endfor*/

       if (ires != 1) {
           for (iatm=1;iatm<=pos->residue[ires-1].natom;iatm++) {
                iiatm     = pos->residue[ires-1].atom[iatm];
                atm_label = maps->upac_atm_label[iiatm];
                if (strcasecmp(atm_label,"CA")==0) {ind_prevres_CA  = iiatm;}
                if (strcasecmp(atm_label,"O")==0)  {ind_prevres_O   = iiatm;}
                if (strcasecmp(atm_label,"CMA")==0){ind_prevres_CMA = iiatm;}
           }/*endfor*/
       }/*endif*/

       if (ires != pos->nres) {
           for (iatm=1;iatm<=pos->residue[ires+1].natom;iatm++) {
                iiatm     = pos->residue[ires+1].atom[iatm];
                atm_label = maps->upac_atm_label[iiatm];
                if (strcasecmp(atm_label,"CA")==0) {ind_nextres_CA  = iiatm;}
                if (strcasecmp(atm_label,"O")==0)  {ind_nextres_O   = iiatm;}
                if (strcasecmp(atm_label,"CMA")==0){ind_nextres_CMA = iiatm;}
           }/*endfor*/
       }/*endif*/

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* III.C.2) Define clbase                                                  */

       if ( (ires % 2) == 0 ) {         
         iclbase++;
         if (iclbase>num_clbase) {
  	     sprintf(err_msg,"Tors_defs:setup_cgresmodel_3pt_cblc:\n"
                             "iclbase>num_clbase");
           errorMessage(err_msg);
         }/*endif*/
         res_id_clbase[iclbase] = ires;
         natm_clbase[iclbase] = 3;
         ind_atm_clbase[iclbase][1] = ind_CMA;         /*CMA*/
         ind_atm_clbase[iclbase][2] = ind_CA;          /*CA*/
         ind_atm_clbase[iclbase][3] = ind_O;           /*O*/
       }/*endif*/

       if (ires == 1) {
         itrans++;
         if (itrans>num_trans) {
 	     sprintf(err_msg,"Tors_defs::setup_cgresmodel_3pt_cblc:\n"
                             "itrans>num_trans");
             errorMessage(err_msg);
         }/*endif*/
         ityp_trans[itrans] = 1;
         prop_sig_scale[itrans] = 0.02; // 0.05 0.1 
         res_id[itrans] = ires;
         natm_trans[itrans] = 3;
         ind_trans[itrans][1] = ind_O;               /*O*/
         ind_trans[itrans][2] = ind_CA;              /*CA*/
         ind_trans[itrans][3] = ind_CMA;             /*CMA*/
         ind_prim_tors[itrans][4] = ind_O;           /*O*/
         ind_prim_tors[itrans][3] = ind_nextres_CMA; /*CMA*/
         ind_prim_tors[itrans][2] = ind_nextres_O;   /*O*/
         ind_prim_tors[itrans][1] = ind_nextres_CA;  /*CA*/
       }/*endif*/

       if ( ((ires % 2) == 1) && (ires != 1) && (ires != (pos->nres)) ) {
         itrans++;
         if (itrans>num_trans) {
 	     sprintf(err_msg,"Tors_defs::setup_cgresmodel_3pt_cblc:\n"
                             "itrans>num_trans");
             errorMessage(err_msg);
         }/*endif*/
         ityp_trans[itrans] = 1;
         prop_sig_scale[itrans] = 0.02; // 0.05 0.1 
         res_id[itrans] = ires;
         natm_trans[itrans] = 2;
         ind_trans[itrans][1] = ind_CA;               /*CA*/
         ind_trans[itrans][2] = ind_CMA;              /*CMA*/
         ind_prim_tors[itrans][4] = ind_CA;           /*CA*/
         ind_prim_tors[itrans][3] = ind_prevres_O;    /*O*/
         ind_prim_tors[itrans][2] = ind_prevres_CA;   /*CA*/
         ind_prim_tors[itrans][1] = ind_prevres_CMA;  /*CMA*/


         itrans++;
         if (itrans>num_trans) {
 	    sprintf(err_msg,"Tors_defs::setup_cgresmodel_3pt_cblc:\n"
                            "itrans>num_trans");
            errorMessage(err_msg);
         }/*endif*/
         ityp_trans[itrans] = 2;          /*needs closure*/
         prop_sig_scale[itrans] = 0.02;
         res_id[itrans] = ires;
         natm_trans[itrans] = 1;
         ind_trans[itrans][1]     = ind_O;          /*O*/
         ind_prim_tors[itrans][4] = ind_O;          /*O*/
         ind_prim_tors[itrans][3] = ind_CA;         /*CA*/
         ind_prim_tors[itrans][2] = ind_prevres_O;  /*O*/
         ind_prim_tors[itrans][1] = ind_prevres_CA; /*CA*/
         /*____________________/closure/____________________*/
         iclos++;
         if (iclos>num_clos) {
    	    sprintf(err_msg,"Tors_defs::setup_cgresmodel_3pt_cblc:\n"
                            "iclos>num_clos");
            errorMessage(err_msg);
         }/*endif*/
         ind_clos_trans[itrans] = iclos;
         ityp_clos[iclos]       = 0;        /*cblc*/
         natm_clos[iclos]       = 5;
         ind_atm_clos[iclos][5] = ind_nextres_CA;   /*CA*/
         ind_atm_clos[iclos][4] = ind_O;            /*O*/
         ind_atm_clos[iclos][3] = ind_CA;           /*CA*/
         ind_atm_clos[iclos][2] = ind_prevres_O;    /*O*/
         ind_atm_clos[iclos][1] = ind_prevres_CA;   /*CA*/
         dx = x[ind_nextres_CA] - x[ind_O];
         dy = y[ind_nextres_CA] - y[ind_O];
         dz = z[ind_nextres_CA] - z[ind_O];
         const_clos[iclos][1] = sqrt(dx*dx+dy*dy+dz*dz); /*r54*/
         dx = x[ind_CA] - x[ind_O];
         dy = y[ind_CA] - y[ind_O];
         dz = z[ind_CA] - z[ind_O];
         const_clos[iclos][2] = sqrt(dx*dx+dy*dy+dz*dz); /*r34*/
       }/*endif*/
  
       if ( (ires == (pos->nres)) && ((ires % 2)!=0)  ) {         
         itrans++;
         if (itrans>num_trans) {
 	     sprintf(err_msg,"Tors_defs::setup_cgresmodel_3pt_cblc:\n"
                             "itrans>num_trans");
             errorMessage(err_msg);
         }/*endif*/
         ityp_trans[itrans] = 1;
         prop_sig_scale[itrans] = 0.02; // 0.05 0.1 
         res_id[itrans] = ires;
         natm_trans[itrans] = 3;
         ind_trans[itrans][1] = ind_CA;              /*CA*/
         ind_trans[itrans][2] = ind_O;               /*O*/
         ind_trans[itrans][3] = ind_CMA;             /*CMA*/
         ind_prim_tors[itrans][4] = ind_CA;          /*CA*/
         ind_prim_tors[itrans][3] = ind_prevres_O;   /*O*/
         ind_prim_tors[itrans][2] = ind_prevres_CA;  /*CA*/
         ind_prim_tors[itrans][1] = ind_prevres_CMA; /*CMA*/
       }/*endif*/
  }/*end for-ires*/
 }/*end if*/


/*-------------------------------------------------------------------------*/
/* III.C) Define the group of atoms to be transformed for nucleic acids    */

 if (nucleic) {
    sprintf(err_msg,"Tors_defs::setup_cgresmodel_3pt_cblc\n"
                    "currently nucleic acids are not supported");
    errorMessage(err_msg);
 }/*end if*/


/*-------------------------------------------------------------------------*/
/* IV) If stride is defined in .pdb, only move basis                       */
/*       along 'C' secondary structure type residue                        */
 
  if (general->stride_defined==1) {
    if (nstride!=1) {
      sprintf(err_msg, "Tors_defs::setup_cgresmodel_5pt_cblc: nstride!=1");
      errorMessage(err_msg);
    }/*endif*/
    if (strlen(&(stride[1][0]))-1!=pos->nres) {
      sprintf(err_msg, "From Tors_defs::setup_cgresmodel_5pt:\n" 
                       "corrupted pos.stride \n");
      errorMessage(err_msg);
    }/*endif*/
    for (iclbase=1;iclbase<=num_clbase;iclbase++) {
         freeze_clbase[iclbase] = 1;
         if (stride[1][res_id_clbase[iclbase]] == 'C') {
	     freeze_clbase[iclbase] = 0;
         }/*endif*/
    }/*endfor*/
  }/*end if*/

/*-------------------------------------------------------------------------*/
/* V) Setup rotate_utils                                                 */

  rotate_utils.setup();
  
/*-------------------------------------------------------------------------*/
/* VI) Debug                                                             */

#ifdef DEBUG
  printf("############################ All Torsions ############################\n");
  write_alldata(maps);
#endif

/*-------------------------------------------------------------------------*/
}/*end routine*/  
/*=========================================================================*/
