#include "../../include/standard.h" 
#include "../../include/typedefs.h"
#include "../../include/definitions.h"
#include "../../include/utillib.h"
#include "../../include/cmalloc.h"

#define DEBUG
#undef DEBUG

/*=========================================================================*/
 void Tors_defs::setup_cgresmodel_5pt(const Maps *maps, const Pos *pos, 
                                                        const General *general)
/*-------------------------------------------------------------------------*/
/*  Sets up all the torsional angles for cgres_model_5pt                   */
/*-------------------------------------------------------------------------*/
{/*begin routine*/  
/*-------------------------------------------------------------------------*/

  int ires, itrans;
  int iatm, iiatm, ibranch;  
  char   err_msg[LINE_NCHARMAX];


  int    nmol  =  maps->nmol;


/*-------------------------------------------------------------------------*/
/* 00) If cblc closure is used                                             */
  if (general->cblc_defined==1) {
      setup_cgresmodel_5pt_cblc(maps,pos,general);
      return;
  }/*endif*/


/*-------------------------------------------------------------------------*/
/* 0) Only one molecule is supported                                       */

  if (nmol!=1) {
    sprintf(err_msg,"Tors_defs::setup_cgresmodel_3pt:nmol!=1:not supported");
    errorMessage(err_msg);
  }/*endif*/

/*-------------------------------------------------------------------------*/
/* I) Get the number of transformations                                    */
/*    Assume two transformations/residue, except for first residue         */

  num_trans = 2*((pos->nres)-1);

/*-------------------------------------------------------------------------*/
/* II) Allocate variables                                                  */

  res_id         =  new int[num_trans]-1;
  ityp_trans     =  new int[num_trans]-1;
  off_trans      =  new int[num_trans]-1;
  natm_trans     =  new int[num_trans]-1;

  ind_trans      =  cmall_int_mat(1, num_trans, 1, NATM_TRANS_MAX);
  ind_prim_tors  =  cmall_int_mat(1, num_trans, 1, 4);


/*-------------------------------------------------------------------------*/
/* III.A)  Decide if 5pt protein or nucleic acid                           */
  int nucleic = 0;
  int protein = 0;
  iatm = pos->residue[1].branch_root_f[0][0];
  if (strcasecmp(maps->upac_atm_label[iatm],"CA")==0) {protein = 1;}  
  if (strcasecmp(maps->upac_atm_label[iatm],"P")==0)  {nucleic = 1;}  
#ifdef DEBUG
      if (protein) {printf("This is a protein in 5pt representation\n");}
      if (nucleic) {printf("This is a nucleic acid in 5pt representation\n");}
      getchar();
#endif

/*-------------------------------------------------------------------------*/
/* III.B) Define the group of atoms to be transformed for proteins         */
 if (protein) {
      sprintf(err_msg, "From Tors_defs::setup_cgresmodel_5pt:\n" 
                       "currently proteins are not supported \n");
      errorMessage(err_msg);
 }/*end if*/
/*-------------------------------------------------------------------------*/
/* III.C) Define the group of atoms to be transformed for nucleic acids   */
/*      1) Loop over the residues having groups (first excluded)           */
/*      2) Assume residues have two branch roots                           */
/*      3) Assume two groups/residue                                       */
/*      4) Group 1 : branch_root_s[0][0]                                   */
/*      5) Group 2 : branch_root_f[0][0] and branch_root_s[i][0] i=1,3     */  
 if (nucleic) {
  itrans = 0;
  for (ires=1;ires<=pos->nres;ires++) {
    if (ires==1) {continue;}
    itrans++;
    res_id[itrans] = ires;
    natm_trans[itrans] = 1;
    ind_trans[itrans][1] = pos->residue[ires].branch_root_s[0][0];
    ind_prim_tors[itrans][4] = pos->residue[ires].branch_root_s[0][0];
    ind_prim_tors[itrans][3] = pos->residue[ires].branch_root_f[0][0];
    ind_prim_tors[itrans][2] = pos->residue[ires-1].branch_root_s[0][0];
    ind_prim_tors[itrans][1] = pos->residue[ires-1].branch_root_f[0][0];

    if (ires!=pos->nres) {
        itrans++;
        res_id[itrans] = ires;
        natm_trans[itrans] = 1;
        ind_trans[itrans][1] = pos->residue[ires+1].branch_root_f[0][0];
        iatm = pos->residue[ires].branch_root_f[0][0];
        for (ibranch=1;ibranch<=pos->residue[ires].nbranch_s1;ibranch++) {
             iiatm =  pos->residue[ires].branch_root_s[ibranch][0];
             if (iiatm != iatm) {
	         natm_trans[itrans]++;
                 ind_trans[itrans][natm_trans[itrans]] = iiatm;
             }/*end if*/
        }/*end for*/
        ind_prim_tors[itrans][4] = pos->residue[ires+1].branch_root_f[0][0];
        ind_prim_tors[itrans][3] = pos->residue[ires].branch_root_s[0][0];
        ind_prim_tors[itrans][2] = pos->residue[ires].branch_root_f[0][0];
        ind_prim_tors[itrans][1] = pos->residue[ires-1].branch_root_s[0][0];
    } else {
        itrans++;
        res_id[itrans] = ires;
        natm_trans[itrans] = 0;
        iatm = pos->residue[ires].branch_root_f[0][0];
        for (ibranch=1;ibranch<=pos->residue[ires].nbranch_s1;ibranch++) {
             iiatm =  pos->residue[ires].branch_root_s[ibranch][0];
             if (iiatm != iatm) {
	         natm_trans[itrans]++;
                 ind_trans[itrans][natm_trans[itrans]] = iiatm;
             }/*end if*/
        }/*end for*/
        ind_prim_tors[itrans][4] = ind_trans[itrans][1];
        ind_prim_tors[itrans][3] = pos->residue[ires].branch_root_s[0][0];
        ind_prim_tors[itrans][2] = pos->residue[ires].branch_root_f[0][0];
        ind_prim_tors[itrans][1] = pos->residue[ires-1].branch_root_s[0][0];
    }
  }/*end for*/
 }/*end if*/
/*-------------------------------------------------------------------------*/
/* IV) Transformation types                                                */
/*     By default, everybody is transformed                                */

  for (itrans=1;itrans<=num_trans;itrans++) {
       ityp_trans[itrans] = 1;
       off_trans[itrans]  = 0;
  }/*end for*/

/*-------------------------------------------------------------------------*/
/* V) If stride is defined in .pdb, only sample                            */
/*       along 'C' secondary structure type residue                        */   

  if (general->stride_defined==1) {
    if (nstride!=1) {
      sprintf(err_msg, "Tors_defs::setup_cgresmodel_5pt: nstride!=1");
      errorMessage(err_msg);
    }/*endif*/
    if (strlen(&(stride[1][0]))-1!=pos->nres) {
      sprintf(err_msg, "From Tors_defs::setup_cgresmodel_5pt:\n" 
                       "corrupted pos.stride \n");
      errorMessage(err_msg);
    }/*endif*/
    for (itrans=1;itrans<=num_trans;itrans++) {
         ityp_trans[itrans] = 1;
         off_trans[itrans]  = 1;
         if (stride[1][res_id[itrans]] == 'C') {
             off_trans[itrans] = 0;
         }/*endif*/
    }/*end for*/
  }/*end if*/


/*-------------------------------------------------------------------------*/
/* VI) Adjust transformation types                                         */
/*     if off_trans == 1 => ityp_trans = 0 (identical)                     */
/*     if off_trans == 0 => ityp_trans = 1 (prop Gauss)                    */

  for (itrans=1;itrans<=num_trans;itrans++) {
    if (off_trans[itrans] == 1) {ityp_trans[itrans] = 0;}
    if (off_trans[itrans] == 0) {ityp_trans[itrans] = 1;}
  }/*end for*/

/*-------------------------------------------------------------------------*/
/* VII) Setup rotate_utils                                                 */

  rotate_utils.setup();
  
/*-------------------------------------------------------------------------*/
/* VIII) Debug                                                             */

#ifdef DEBUG
  printf("############################ All Torsions ############################\n");
  write_alldata(maps);
#endif

/*-------------------------------------------------------------------------*/
}/*end routine*/  
/*=========================================================================*/
