#include "../../../include/standard.h" 
#include "../../../include/typedefs.h"
#include "../../../include/definitions.h"
#include "../../../include/utillib.h"
#include "../../../include/cmalloc.h"

#define DEBUG
#undef DEBUG

/*=========================================================================*/
 void Stage_defs::setup_model_AA(const Maps *maps,
                                 const Pos *pos,
                                 const General *general)
/*-------------------------------------------------------------------------*/
/*     Sets up all the torsional angles for All Atom model                 */
/*-------------------------------------------------------------------------*/
{/*begin routine*/
/*-------------------------------------------------------------------------*/

  int i, imol, nres;
  int ires, iresl, itrans;
  int stage_supported;

  char chainid;

  int    nmol        = maps->nmol;
  int   *nres_mol    = maps->nres_mol;
  int   *npro_mol    = maps->npro_mol;
  int  **ind_res_mol = maps->ind_res_mol;
  char **mol_label   = maps->mol_label;
  char **mol_chainid = maps->mol_chainid;

  char   err_msg[LINE_NCHARMAX];

/*-------------------------------------------------------------------------*/
/* 0) Get information about molecule types                                 */
/*     (protein, DNA, RNA, water, atom/ion)                                */
/*      Currently only protein is supported                                */

     for (imol=1;imol<=nmol;imol++) {
       stage_supported = 0;
       if (strcasecmp(mol_label[imol],"protein")==0) {stage_supported=1;}
       if (strcasecmp(mol_label[imol],"DNA")==0)     {stage_supported=1;}
          if (stage_supported == 0) {
              sprintf(err_msg,"From Stage_defs::setup_model_AA           \n"
                             " Currenlty only protein molecules/chains   \n"
                             " and DNA strands/chains are supported        "
		             " with \\sym_type{PTSTG/PTSTGMC}              ");
              errorMessage(err_msg);
          }/*endif*/
     }/*endfor*/

/*-------------------------------------------------------------------------*/
/* I) Get the number of transformations, natm_trans_max                    */
/*                                                                         */
/*    A) PROTEINS                                                          */
/*       Backbone: Assume three transformations per residue                */
/*                        six transformations/proline residues             */
/*                        plus one transformation for main atomic backbone */  
/*       Side Chain torsions: Get the number of chi groups from topology   */
/*                 Assume one transformation per every chi group           */
/*                 plus one transformation for side chain backbone         */  
/*    B) DNA                                                               */
/*       Backbone: Assume seven transformations per nucleotide             */
/*                        plus one transformation for main atomic backbone */  
/*       Base torsions: Assume Nb transformations for different bases      */
/*                 Nb(ADE) = 6, Nb(THY) = 8, Nb(CYT) = 7, Nb(GUA) = 7      */

     num_trans  = 0;
     natm_trans_max = 0;
     for (imol=1;imol<=nmol;imol++) {
          if (strcasecmp(mol_label[imol],"protein")==0) {

              nres = nres_mol[imol];
              natm_trans_max = MAX(natm_trans_max,3*nres); 
              num_trans += 3*(nres + npro_mol[imol]) + 1; /*Protein Backbone*/

              for (iresl=1;iresl<=nres;iresl++) {
                   ires = ind_res_mol[imol][iresl];
                   num_trans += pos->residue[ires].nchigrp;      /*Sidechain*/
                   if (pos->residue[ires].nchigrp > 0) {
		       num_trans += 1;
                   }/*end if*/
              }/*end for*/
           }/*endif*/

           if (strcasecmp(mol_label[imol],"DNA")==0) {
              nres = nres_mol[imol];
              natm_trans_max = MAX(natm_trans_max,6*nres); 
              num_trans += 7*(nres) + 1;                      /*DNA Backbone*/
              for (iresl=1;iresl<=nres;iresl++) {
                   ires = ind_res_mol[imol][iresl];
                   if (strcasecmp(pos->residue[ires].residue_name,"ADE")==0)
		     {num_trans += 6;}   
                   if (strcasecmp(pos->residue[ires].residue_name,"THY")==0)
		     {num_trans += 8;}   
                   if (strcasecmp(pos->residue[ires].residue_name,"CYT")==0)
		     {num_trans += 7;}   
                   if (strcasecmp(pos->residue[ires].residue_name,"GUA")==0)
		     {num_trans += 7;}
              }/*end for*/
           }/*endif*/

     }/*endfor-imol*/


/*-------------------------------------------------------------------------*/
/* II) Allocate variables                                                  */

  res_id         =  new int[num_trans]-1;
  ityp_trans     =  new int[num_trans]-1;
  natm_trans     =  new int[num_trans]-1;

  ind_trans      =  cmall_int_mat(1, num_trans, 1, natm_trans_max);

/*-------------------------------------------------------------------------*/
/* III) Setup torsional transformation for molecules                       */
/*      Subsequently check for inconsistencies                             */

  itrans  = 0;


#ifdef DEBUG
  printf("Check Point I\n");
  getchar();
#endif


  for (imol=1;imol<=nmol;imol++) {
      if (strcasecmp(mol_label[imol],"protein")==0) {
	   setup_AA_protein(maps,pos,general,imol,&itrans);
      }/*endif*/
      if (strcasecmp(mol_label[imol],"DNA")==0) {
	   setup_AA_DNA(maps,pos,general,imol,&itrans);
      }/*endif*/
  }/*endfor*/


  if (itrans!=num_trans) {
      sprintf(err_msg,"Stage_defs::setup_model_AA -> itrans != num_trans\n"
                      "itrans = %d, num_trans = %d",itrans,num_trans);
      errorMessage(err_msg);
  }/*endif*/
  
/*-------------------------------------------------------------------------*/
/* IX) Debug                                                               */


#ifdef DEBUG
  printf("~~~~~~~~~~~~~~~~~~~~~~~~~ All Stage ~~~~~~~~~~~~~~~~~~~~~~~~~\n");
  write_alldata(maps);
#endif

/*-------------------------------------------------------------------------*/
}/*end routine*/  
/*=========================================================================*/

