#include "../include/standard.h" 
#include "../include/typedefs.h"
#include "../include/definitions.h"
#include "../include/utillib.h"
#include "../include/cmalloc.h"

#define DEBUG_OFF


/*-------------------------------------------------------------------------*/
/*                        closure algorithms                               */
/* return 1 => successful closure                                          */
/* return 0 => non successful closure                                      */
/* ityp_clos ==  0 : cblc closure (J. Comp. Chem. 27 309-315 (2006))       */
/*                   there are 5 atoms involved                            */
/* ityp_clos ==  1 :                                                       */
/*-------------------------------------------------------------------------*/
/*ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc*/
/*-------------------------------------------------------------------------*/

int closure(Pos *pos,Tors_defs *tors_defs,int i_trans,int type) 

/*-------------------------------------------------------------------------*/
  {/*begin*/
/*-------------------------------------------------------------------------*/

  int j1,j2,j3,j4,j5;
  int iclos;
  int i,j;

  int ndim, call_closure;
  double norm, norm_e1, norm_e2, eps;
  double scl,scl_2,scl_p,scl_m;
  double omega3_targ, omega3_plus, omega3_minus;
  double sqrtarg1, sqrtarg2;
  double r45_curr, r34_curr;
  double r34_2, r35_2, r45_2;  
  double dx, dy, one;
  double xx, yy, p;
  double arg;
  double randp10; 

  double *ux  = pos->xi;
  double *uy  = pos->yi;
  double *uz  = pos->zi;

  double*  r1  =  tors_defs->rotate_utils.r1;
  double*  r2  =  tors_defs->rotate_utils.r2;
  double*  r3  =  tors_defs->rotate_utils.r3;
  double*  r4  =  tors_defs->rotate_utils.r4;
  double*  r5  =  tors_defs->rotate_utils.r4_;

  double*  e1  =  tors_defs->rotate_utils.e1;
  double*  e2  =  tors_defs->rotate_utils.e2;
  double*  ex  =  tors_defs->rotate_utils.ex;
  double*  ey  =  tors_defs->rotate_utils.ey;
  double*  ez  =  tors_defs->rotate_utils.ez;

  int *ind_clos_trans = tors_defs->ind_clos_trans;
  int *ityp_clos      = tors_defs->ityp_clos;
  int *natm_clos      = tors_defs->natm_clos;
  int **ind_atm_clos  = tors_defs->ind_atm_clos;
  double **const_clos = tors_defs->const_clos;

  extern  System  *global_system;
  System *system = global_system;
  long *seed     = &(system->general.seed);
  double sig_move =  system->general.prop_clos_sig;
                    

  char   err_msg[LINE_NCHARMAX];

/*-------------------------------------------------------------------------*/
/* I) locate closure (get the closure index) and define separate           */
/*    closure algorithms                                                   */
/*-------------------------------------------------------------------------*/
/* A) ityp_clos == 0                                                       */
/*    The goal of this algorithm is to find a new position for atom 4 such */
/*    that it stisfies the constant bond length approximation, namely      */
/*    |r34| = d(|r3-r4|) and |r45| = d(|r4-r5|)                            */
/*                5-4-3-2-1 : ez = (r23 x r34)/|r23 x r34|                 */
/*                          : ex = (r35 - p ez)/|r35 - p ez|               */
/*                          : p  = ez . r35                                */
/*                          : ey = ez x ex                                 */
/*                                                                         */
/*      indices of closure atoms: (j1,j2,j3,j4,j5)                         */
/*      assign closure atoms    : (r1,r2,r3,r4,r5)                         */
/*    closure algorithms                                                   */
/*-------------------------------------------------------------------------*/
/* B) ityp_clos == 1,2,3,4                                                 */
/*    Using the closure of A) this closure is trying to find the position  */
/*    of atom 4, but it does not exit if the geometric conditins are not   */
/*    satisfied, but it solves only half of the closure (r45) and calls    */
/*    a 'new' closure problem for a subsequent lower level                 */
/*    The following possibilities can arise:                               */
/*                                                                         */
/*    1: same as A) but in case of geometric failure it moves atom 4       */
/*       towards atom 5 so that the r45 is satisfies the bond constraint   */
/*       then submits the solution to a lower level                        */
/*    2: random displacement of atom 4 towards atom 5 and random growth of */
/*       the r34 bond then trying to use A) to solve the 'new' problem     */
/*       if the later fails use the same recovery provided in 1 and        */
/*       in case of success it calls the closure relation for a lower level*/   /*    3: always move atom 4 towards atom 5, so that r45 is satisfied       */
/*       and calls the closure soltion for a lower level                   */
/*    4: 2 without the random dispacement of atom 4 towards atom 5         */
/*    5: apply type 3 then make a random displacement of r4 on the surface */
/*       of the sphere (radius is r45) around r5                           */
/*                                                                         */
/*    (r45_2 > p*p)  geometric failure                                     */
/*    (p^2 == r35_2) geometric failure                                     */
/*    (r34_2 < x*x)  geometric failure                                     */
/*    where x, p are defined below                                         */
/*-------------------------------------------------------------------------*/

  ndim  = 3;

  if (ind_clos_trans[i_trans]==0) {return 0;}
  iclos = ind_clos_trans[i_trans];

#ifdef DEBUG
      printf("ityp_clos[%d] %d\n",iclos,ityp_clos[iclos]);
      pos->write_tors_defs(*tors_defs,i_trans);
      getchar();
#endif

      r45_2  = const_clos[iclos][1];    /* bond constraint |r45|  */
      r45_2 *= const_clos[iclos][1];

      r34_2  = const_clos[iclos][2];    /* bond constraint |r34|  */
      r34_2 *= const_clos[iclos][2];

      j1 = ind_atm_clos[iclos][1];
      j2 = ind_atm_clos[iclos][2];
      j3 = ind_atm_clos[iclos][3];
      j4 = ind_atm_clos[iclos][4];
      j5 = ind_atm_clos[iclos][5];

      r1[1] = ux[j1]; r1[2] = uy[j1]; r1[3] = uz[j1];
      r2[1] = ux[j2]; r2[2] = uy[j2]; r2[3] = uz[j2];
      r3[1] = ux[j3]; r3[2] = uy[j3]; r3[3] = uz[j3];
      r4[1] = ux[j4]; r4[2] = uy[j4]; r4[3] = uz[j4];
      r5[1] = ux[j5]; r5[2] = uy[j5]; r5[3] = uz[j5];

      r45_curr =   sqrt((r4[1]-r5[1])*(r4[1]-r5[1]) + 
                        (r4[2]-r5[2])*(r4[2]-r5[2]) +
	                (r4[3]-r5[3])*(r4[3]-r5[3]));
#ifdef DEBUG
      printf("before closure\n");
      printf("r4x %g, r4y %g, r4z %g\n",r4[1],r4[2],r4[3]);
      printf("r5x %g, r5y %g, r5z %g\n",r5[1],r5[2],r5[3]);
      printf("r45_curr %g\n",r45_curr);
      printf("r45_targ %g\n",sqrt(r45_2));
#endif      

      /*make a random move towards r5 and a random scaling to r34*/
      scl = 1.0;
      if (ityp_clos[iclos] == 2) {	       
                randp10 = ran1(seed)/10.0;
                e1[1] = r4[1]-r5[1]; 
                e1[2] = r4[2]-r5[2]; 
                e1[3] = r4[3]-r5[3]; /*r54*/
                norm_e1 =  sqrt(e1[1]*e1[1]+e1[2]*e1[2]+e1[3]*e1[3]);
                for (i=1;i<=ndim;i++){e1[i] /= norm_e1;}
                r4[1] = r5[1] + (1.0+randp10)*sqrt(r45_2)*e1[1];
                r4[2] = r5[2] + (1.0+randp10)*sqrt(r45_2)*e1[2];
                r4[3] = r5[3] + (1.0+randp10)*sqrt(r45_2)*e1[3];
		scl   = sqrt(randp10);
                scl_2 = randp10;
                r34_2 *= scl_2;                 
      }/*endif*/

      if (ityp_clos[iclos] == 4) {
                randp10 = ran1(seed)/10.0;
                scl   = sqrt(randp10);
                scl_2 = randp10;
                r34_2 *= scl_2;                 
      }/*endif*/


      /* omega3-bend angle */ 
      e1[1] = r3[1]-r2[1]; e1[2] = r3[2]-r2[2]; e1[3] = r3[3]-r2[3];  /*r23*/
      e2[1] = r4[1]-r3[1]; e2[2] = r4[2]-r3[2]; e2[3] = r4[3]-r3[3];  /*r34*/
      norm_e1 =  sqrt(e1[1]*e1[1]+e1[2]*e1[2]+e1[3]*e1[3]);
      for (i=1;i<=ndim;i++){e1[i] /= norm_e1;}
      norm_e2 =  sqrt(e2[1]*e2[1]+e2[2]*e2[2]+e2[3]*e2[3]);
      for (i=1;i<=ndim;i++){e2[i] /= norm_e2;}
      arg =  (e1[1]*e2[1]+e1[2]*e2[2]+e1[3]*e2[3]);
      omega3_targ = acos(arg);

      /* ez */
      cprd(e1[1],e1[2],e1[3],e2[1],e2[2],e2[3],&(ez[1]),&(ez[2]),&(ez[3]));
      norm =  sqrt(ez[1]*ez[1]+ez[2]*ez[2]+ez[3]*ez[3]);
      for (i=1;i<=ndim;i++){ez[i] /= norm;}

      /*r35_2*/
      e1[1] = r5[1]-r3[1]; e1[2] = r5[2]-r3[2]; e1[3] = r5[3]-r3[3];  /*r35*/
      r35_2 =  e1[1]*e1[1]+e1[2]*e1[2]+e1[3]*e1[3];

      /* ex */
      p =  e1[1]*ez[1] + e1[2]*ez[2] + e1[3]*ez[3];
      ex[1] = e1[1] - p*ez[1];
      ex[2] = e1[2] - p*ez[2];
      ex[3] = e1[3] - p*ez[3];
      norm =  sqrt(ex[1]*ex[1]+ex[2]*ex[2]+ex[3]*ex[3]);
      for (i=1;i<=ndim;i++){ex[i] /= norm;}
      
      /* ey */
      cprd(ez[1],ez[2],ez[3],ex[1],ex[2],ex[3],&(ey[1]),&(ey[2]),&(ey[3]));

      /* xx, yy */
      xx = (r34_2 + r35_2 - r45_2)/(2.0*sqrt(r35_2-p*p));
      yy = (sqrt(r34_2-xx*xx));

      /*unwanted geometries*/
      eps = 1.e-13;
      if ((p*p < r35_2 + eps) && (p*p > r35_2 - eps)) {
	  return 0;
      }/*endif*/
#ifdef DEBUG
      if (p*p > r45_2) {
          printf("p*p > r45_2\n");
          getchar();
      }/*endif*/
#endif

      if (ityp_clos[iclos] == 0) {
	   if ((r34_2 < xx*xx) || (r34_2 < xx*xx)) {return 0;}
      }
      else if (ityp_clos[iclos] == 1) {
           if ( (p*p > r45_2) || (r34_2 < xx*xx) ) {
                e1[1] = r4[1]-r5[1]; 
                e1[2] = r4[2]-r5[2]; 
                e1[3] = r4[3]-r5[3]; /*r54*/
                norm_e1 =  sqrt(e1[1]*e1[1]+e1[2]*e1[2]+e1[3]*e1[3]);
                for (i=1;i<=ndim;i++){e1[i] /= norm_e1;}
                r4[1] = r5[1] + sqrt(r45_2)*e1[1];
                r4[2] = r5[2] + sqrt(r45_2)*e1[2];
                r4[3] = r5[3] + sqrt(r45_2)*e1[3];
                ux[j4] = r4[1]; uy[j4] = r4[2]; uz[j4] = r4[3];
	        return closure(pos,tors_defs,i_trans-1,-1);
          }/*endif*/
      }
      else if ((ityp_clos[iclos] == 2) || (ityp_clos[iclos] == 4)) {
           if ( (p*p > r45_2) || (r34_2 < xx*xx) ) {
                e1[1] = r4[1]-r5[1]; 
                e1[2] = r4[2]-r5[2]; 
                e1[3] = r4[3]-r5[3]; /*r54*/
                norm_e1 =  sqrt(e1[1]*e1[1]+e1[2]*e1[2]+e1[3]*e1[3]);
                for (i=1;i<=ndim;i++){e1[i] /= norm_e1;}
                r4[1] = r5[1] + sqrt(r45_2)*e1[1];
                r4[2] = r5[2] + sqrt(r45_2)*e1[2];
                r4[3] = r5[3] + sqrt(r45_2)*e1[3];
                ux[j4] = r4[1]; uy[j4] = r4[2]; uz[j4] = r4[3];
	        return closure(pos,tors_defs,i_trans-1,-1);
           }/*endif*/
      }
      else if (ityp_clos[iclos] == 3) {
                e1[1] = r4[1]-r5[1]; 
                e1[2] = r4[2]-r5[2]; 
                e1[3] = r4[3]-r5[3]; /*r54*/
                norm_e1 =  sqrt(e1[1]*e1[1]+e1[2]*e1[2]+e1[3]*e1[3]);
                for (i=1;i<=ndim;i++){e1[i] /= norm_e1;}
                r4[1] = r5[1] + sqrt(r45_2)*e1[1];
                r4[2] = r5[2] + sqrt(r45_2)*e1[2];
                r4[3] = r5[3] + sqrt(r45_2)*e1[3];
                ux[j4] = r4[1]; uy[j4] = r4[2]; uz[j4] = r4[3];
	        return closure(pos,tors_defs,i_trans-1,-1);
      }
      else if (ityp_clos[iclos] == 5) {
                e1[1] = r4[1]-r5[1]; 
                e1[2] = r4[2]-r5[2]; 
                e1[3] = r4[3]-r5[3]; /*r54*/
                norm_e1 =  sqrt(e1[1]*e1[1]+e1[2]*e1[2]+e1[3]*e1[3]);
                for (i=1;i<=ndim;i++){e1[i] /= norm_e1;}
                r4[1] = r5[1] + sqrt(r45_2)*e1[1];
                r4[2] = r5[2] + sqrt(r45_2)*e1[2];
                r4[3] = r5[3] + sqrt(r45_2)*e1[3];
                ex[1] = ran1(seed); ex[2] = ran1(seed);           
                ex[3] = -(ex[1]*e1[1] + ex[2]*e1[2])/e1[3];      /*ex _|_ e1*/
                norm =  sqrt(ex[1]*ex[1]+ex[2]*ex[2]+ex[3]*ex[3]);
                for (i=1;i<=ndim;i++){ex[i] /= norm;}
                cprd(e1[1],e1[2],e1[3],                   /*ey _|_ ex _|_ e1*/
                     ex[1],ex[2],ex[3],&(ey[1]),&(ey[2]),&(ey[3]));
                one = 1.0;
                dx = mb(one,sig_move,seed);        /* move on the surface of*/
                dy = mb(one,sig_move,seed);        /* the sphere around r5  */
                r4[1] = r4[1] + dx*ex[1] + dy*ey[1];
                r4[2] = r4[2] + dx*ex[2] + dy*ey[2];
                r4[3] = r4[3] + dx*ex[3] + dy*ey[3];
                e1[1] = r4[1]-r5[1]; 
                e1[2] = r4[2]-r5[2]; 
                e1[3] = r4[3]-r5[3]; /*r54*/
                norm_e1 =  sqrt(e1[1]*e1[1]+e1[2]*e1[2]+e1[3]*e1[3]);
                for (i=1;i<=ndim;i++){e1[i] /= norm_e1;}
                r4[1] = r5[1] + sqrt(r45_2)*e1[1];
                r4[2] = r5[2] + sqrt(r45_2)*e1[2];
                r4[3] = r5[3] + sqrt(r45_2)*e1[3];

                ux[j4] = r4[1]; uy[j4] = r4[2]; uz[j4] = r4[3];
	        return closure(pos,tors_defs,i_trans-1,-1);
      }/*endif*/


      /* closure (-) */
      r4[1] = r3[1] + xx*ex[1] - yy*ey[1]; /* or +yy*ey[1] */
      r4[2] = r3[2] + xx*ex[2] - yy*ey[2]; /* or +yy*ey[2] */ 
      r4[3] = r3[3] + xx*ex[3] - yy*ey[3]; /* or +yy*ey[3] */ 

      /* omega3_minus */ 
      e1[1] = r3[1]-r2[1]; e1[2] = r3[2]-r2[2]; e1[3] = r3[3]-r2[3];  /*r23*/
      e2[1] = r4[1]-r3[1]; e2[2] = r4[2]-r3[2]; e2[3] = r4[3]-r3[3];  /*r34*/
      norm_e1 =  sqrt(e1[1]*e1[1]+e1[2]*e1[2]+e1[3]*e1[3]);
      for (i=1;i<=ndim;i++){e1[i] /= norm_e1;}
      norm_e2 =  sqrt(e2[1]*e2[1]+e2[2]*e2[2]+e2[3]*e2[3]);
      for (i=1;i<=ndim;i++){e2[i] /= norm_e2;}
      arg =  (e1[1]*e2[1]+e1[2]*e2[2]+e1[3]*e2[3]);
      omega3_minus = acos(arg);

      /* save closure (-) */
      ux[j4] = r4[1]; uy[j4] = r4[2]; uz[j4] = r4[3];

      /* closure (+) */
      r4[1] = r3[1] + xx*ex[1] + yy*ey[1]; /* or -yy*ey[1] */
      r4[2] = r3[2] + xx*ex[2] + yy*ey[2]; /* or -yy*ey[2] */ 
      r4[3] = r3[3] + xx*ex[3] + yy*ey[3]; /* or -yy*ey[3] */ 

      /* omega3_plus */ 
      e1[1] = r3[1]-r2[1]; e1[2] = r3[2]-r2[2]; e1[3] = r3[3]-r2[3];  /*r23*/
      e2[1] = r4[1]-r3[1]; e2[2] = r4[2]-r3[2]; e2[3] = r4[3]-r3[3];  /*r34*/
      norm_e1 =  sqrt(e1[1]*e1[1]+e1[2]*e1[2]+e1[3]*e1[3]);
      for (i=1;i<=ndim;i++){e1[i] /= norm_e1;}
      norm_e2 =  sqrt(e2[1]*e2[1]+e2[2]*e2[2]+e2[3]*e2[3]);
      for (i=1;i<=ndim;i++){e2[i] /= norm_e2;}
      arg = (e1[1]*e2[1]+e1[2]*e2[2]+e1[3]*e2[3]);
      omega3_plus = acos(arg);

      /* decide if + is better */
      if ( fabs(omega3_plus-omega3_targ) < fabs(omega3_minus-omega3_targ) ) {
           ux[j4] = r4[1]; uy[j4] = r4[2]; uz[j4] = r4[3];
      } else {
	   r4[1] = ux[j4]; r4[2] = uy[j4]; r4[3] = uz[j4];
      }/*endif*/

#ifdef DEBUG
      printf("after closure\n");
      printf("r4x %g, r4y %g, r4z %g\n",r4[1],r4[2],r4[3]);
      printf("r5x %g, r5y %g, r5z %g\n",r5[1],r5[2],r5[3]);
      r45_curr =   sqrt((r4[1]-r5[1])*(r4[1]-r5[1]) + 
                        (r4[2]-r5[2])*(r4[2]-r5[2]) +
	                (r4[3]-r5[3])*(r4[3]-r5[3]));
      printf("r45_curr %g\n",r45_curr);
      printf("r45_targ %g\n",sqrt(r45_2));
      r34_curr =   sqrt((r3[1]-r4[1])*(r3[1]-r4[1]) + 
                        (r3[2]-r4[2])*(r3[2]-r4[2]) +
	                (r3[3]-r4[3])*(r3[3]-r4[3]));
      printf("r34_curr %g\n",r34_curr);
//      printf("r34_targ*scale(%g) %g\n",scl,scl*sqrt(r34_2));
      printf("r34_targ (%g) %g\n",sqrt(r34_2));
      printf("omega3_targ %g\n",omega3_targ);
      printf("omega3_plus %g\n",omega3_plus);
      printf("omega3_minus %g\n",omega3_minus);
      getchar();
#endif

      if (ityp_clos[iclos]==2) {
 	  return closure(pos,tors_defs,i_trans-1,-1);
      }/*endif*/


      if (ityp_clos[iclos]==4) {
 	  return closure(pos,tors_defs,i_trans-1,-1);
      }/*endif*/

/*-------------------------------------------------------------------------*/
/* II) DEBUG                                                               */

  return 1;

/*-------------------------------------------------------------------------*/
}/*end of closure*/
/*-------------------------------------------------------------------------*/

